from __future__ import annotations

import os
import inspect
from typing import TYPE_CHECKING, Any, Type, Tuple, Union, Generic, TypeVar, Callable, Optional, cast
from datetime import date, datetime
from typing_extensions import (
    Unpack,
    Literal,
    ClassVar,
    Protocol,
    Required,
    Sequence,
    ParamSpec,
    TypedDict,
    TypeGuard,
    final,
    override,
    runtime_checkable,
)

import pydantic
from pydantic.fields import FieldInfo
from dataclasses import is_dataclass

from openai._types import (
    Body,
    IncEx,
    Query,
    ModelT,
    Headers,
    Timeout,
    NotGiven,
    AnyMapping,
    HttpxRequestFiles,
)
from openai._utils import (
    PropertyInfo,
    is_list,
    is_given,
    json_safe,
    lru_cache,
    is_mapping,
    parse_date,
    coerce_boolean,
    parse_datetime,
    strip_not_given,
    extract_type_arg,
    is_annotated_type,
    is_type_alias_type,
    strip_annotated_type,
)
from openai._compat import (
    PYDANTIC_V2,
    ConfigDict,
    GenericModel as BaseGenericModel,
    get_args,
    is_union,
    parse_obj,
    get_origin,
    is_literal_type,
    get_model_config,
    get_model_fields,
    field_get_default,
)
from openai._constants import RAW_RESPONSE_HEADER

from openai._models import (
    validate_type, _build_discriminated_union_meta, BaseModel, GenericModel
)


def hepai_construct_type(*, value: object, type_: object) -> object:
    """Loose coercion to the expected type with construction of nested values.

    If the given value does not match the expected type then it is returned as-is.
    """

    # store a reference to the original type we were given before we extract any inner
    # types so that we can properly resolve forward references in `TypeAliasType` annotations
    original_type = None

    # we allow `object` as the input type because otherwise, passing things like
    # `Literal['value']` will be reported as a type error by type checkers
    type_ = cast("type[object]", type_)
    if is_type_alias_type(type_):
        original_type = type_  # type: ignore[unreachable]
        type_ = type_.__value__  # type: ignore[unreachable]

    # unwrap `Annotated[T, ...]` -> `T`
    if is_annotated_type(type_):
        meta: tuple[Any, ...] = get_args(type_)[1:]
        type_ = extract_type_arg(type_, 0)
    else:
        meta = tuple()

    # we need to use the origin class for any types that are subscripted generics
    # e.g. Dict[str, object]
    origin = get_origin(type_) or type_
    args = get_args(type_)

    if is_union(origin):
        try:
            return validate_type(type_=cast("type[object]", original_type or type_), value=value)
        except Exception:
            pass

        # if the type is a discriminated union then we want to construct the right variant
        # in the union, even if the data doesn't match exactly, otherwise we'd break code
        # that relies on the constructed class types, e.g.
        #
        # class FooType:
        #   kind: Literal['foo']
        #   value: str
        #
        # class BarType:
        #   kind: Literal['bar']
        #   value: int
        #
        # without this block, if the data we get is something like `{'kind': 'bar', 'value': 'foo'}` then
        # we'd end up constructing `FooType` when it should be `BarType`.
        discriminator = _build_discriminated_union_meta(union=type_, meta_annotations=meta)
        if discriminator and is_mapping(value):
            variant_value = value.get(discriminator.field_alias_from or discriminator.field_name)
            if variant_value and isinstance(variant_value, str):
                variant_type = discriminator.mapping.get(variant_value)
                if variant_type:
                    return hepai_construct_type(type_=variant_type, value=value)

        # if the data is not valid, use the first variant that doesn't fail while deserializing
        for variant in args:
            try:
                return hepai_construct_type(value=value, type_=variant)
            except Exception:
                continue

        raise RuntimeError(f"Could not convert data into a valid instance of {type_}")

    if origin == dict:
        if not is_mapping(value):
            return value

        _, items_type = get_args(type_)  # Dict[_, items_type]
        return {key: hepai_construct_type(value=item, type_=items_type) for key, item in value.items()}

    if (
        not is_literal_type(type_)
        and inspect.isclass(origin)
        and (issubclass(origin, BaseModel) or issubclass(origin, GenericModel))
    ):
        if is_list(value):
            return [cast(Any, type_).construct(**entry) if is_mapping(entry) else entry for entry in value]

        if is_mapping(value):
            if issubclass(type_, BaseModel):
                return type_.construct(**value)  # type: ignore[arg-type]

            return cast(Any, type_).construct(**value)

    if origin == list:
        if not is_list(value):
            return value

        inner_type = args[0]  # List[inner_type]
        return [hepai_construct_type(value=entry, type_=inner_type) for entry in value]

    if origin == float:
        if isinstance(value, int):
            coerced = float(value)
            if coerced != value:
                return value
            return coerced

        return value
    
    if is_dataclass(origin):  # ddf2修改，支持dataclass
        return origin(**value)

    if type_ == datetime:
        try:
            return parse_datetime(value)  # type: ignore
        except Exception:
            return value

    if type_ == date:
        try:
            return parse_date(value)  # type: ignore
        except Exception:
            return value

    return value



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from .._models import BaseModel
from .moderation import Moderation

__all__ = ["ModerationCreateResponse"]


class ModerationCreateResponse(BaseModel):
    id: str
    """The unique identifier for the moderation request."""

    model: str
    """The model used to generate the moderation results."""

    results: List[Moderation]
    """A list of moderation objects."""



import os, sys
from pathlib import Path
pydir = Path(os.path.abspath(__file__)).parent

# from damei.nn.api import MODULES, SCRIPTS, IOS
# from damei.nn.api import Config
# from damei.nn.api import AbstractInput, AbstractOutput, AbstractModule
from hai import MODULES, SCRIPTS, IOS, Config
from hai import AbstractModule, AbstractInput, AbstractOutput, AbstractQue

from .dataloader import Dataloader


@IOS.register_module(name='ImagesLoader')
class VisLoader(AbstractInput):
    name = 'ImagesLoader'
    description = '图像数据加载器(支持：image, video, folds, rtsp, rtmp等) hai'
    module_path = f'{Path(os.path.abspath(__file__))}'

    def __init__(self, **kwargs):
        """
        Visible light data loader, such as image, video, folds, rtsp, rtmp etc.
        :param kwargs:
            source: data source, path or url
            img_size: image size, default is 640
        :return:
            4 element tuple of ecah entity: (img_path, img, raw_img, is_cap)
        example:
        >>> vis_loader = VisLoader(source='/path/to/data', img_size=640)
        >>> for i, (path, img, im0, is_cap) in enumerate(vis_loader.data):
        >>>     print(i, path, img.shape, im0.shape, is_cap)
        """
        super(VisLoader, self).__init__(**kwargs)

        source = kwargs.pop('source', None)
        img_size = kwargs.pop('img_size', 640)
        batch_size = kwargs.pop('batch_size', 1)

        if isinstance(source, str) and batch_size:
            source = [source]

        self.data = Dataloader(source=source, imgsz=img_size)


if __name__ == '__main__':
    visloader = VisLoader()



"""
functions of hai datasets hub
"""
import os, sys
from pathlib import Path
pydir = Path(os.path.abspath(__file__)).parent

import json
import shutil
import damei as dm


from .dataset_utils import get_file, extract_archive
from ..utils import general

class DatasetsHub(object):
    def __init__(self) -> None:
        self.urls = f'{pydir.parent.parent}/configs/urls/datasets.json'
        self._dataset_names = None
        self._datasets_dict = None
        self._init()

    def _init(self):
        with open(self.urls, 'r') as f:
            data = json.load(f)
        
        self._datasets_dict = data
        self._dataset_names = list(self._datasets_dict.keys())

    @property
    def names(self):
        return self._dataset_names

    @property
    def datasets(self):
        return self._datasets_dict

    @property
    def datasets_dict(self):
        return self._datasets_dict
    

    def __call__(self, *args, **kwargs):
        pass

    def list(self, **kwargs):
        ret_fmt = kwargs.get('ret_fmt', None)
        if ret_fmt is None:
            info = dm.misc.dict2info(self.datasets)
        elif ret_fmt == 'json':
            info = json.dumps(self.datasets, indent=4)
        elif ret_fmt == 'dict':
            info = self.datasets
        else:
            raise ValueError(f'Unsupported ret_fmt: {ret_fmt}')
        return info

    def list_remote_datasets(self, **kwargs):
        return self.list(**kwargs)

    def list_local_datasets(self, **kwargs):
        ret_fmt = kwargs.get('ret_fmt', 'dict')
        from ...apis import hai_config
        dset_root = hai_config.DATASETS_ROOT
        dsets = [x for x in os.listdir(dset_root) if os.path.isdir(os.path.join(dset_root, x))]
        if ret_fmt == 'list':
            return dsets
        elif ret_fmt == 'dict':
            ret_dict = {f'Total {len(dsets)} datasets': dsets}
            return ret_dict
        else:
            raise ValueError(f'Unsupported ret_fmt: {ret_fmt}')
    

    def download(self, *args, **kwargs):
        name = kwargs.get('name', None)
        version = kwargs.get('version', None)
        save_dir = kwargs.get('save_dir', '.')
        extract = kwargs.get('extract', False)
        force_download = kwargs.get('force_download', False)
        hash_algorithm = kwargs.get('hash_algorithm', 'md5')
        archive_format = kwargs.get('archive_format', 'auto')

        assert name is not None, 'dataset name is required, like: "hai datasets download SFID:latest"'
        name = name.lower()
        if ":" in name:
            name, version = name.split(":")
        version = version if version else 'latest'

        version = general.latest2determined(self.datasets_dict, name=name) if version == 'latest' else version

        # macth dataset by name and version
        data_entry = general.dict_match_item(self.datasets_dict, name=name, version=version)
        # print('entry: ', data_entry)
        origin = data_entry['url']
        fname = data_entry['name']
        file_hash = data_entry['md5'] if 'md5' in data_entry else None
        
        f_path, success = get_file(
            origin=origin,
             fname=fname,
             file_hash=file_hash,
             datadir=save_dir,
             hash_algorithm=hash_algorithm,
             extract=extract,
             force_download=force_download,
             archive_format=archive_format
           ) 
        return f_path, success
        
    def download_dataset(dataset, tp, envfile, force_download):
        """deprecated"""
        # info = datasets[dataset]
        basedir = tp
        datadir = os.path.join(basedir, dataset)
        if force_download:
            if os.path.exists(datadir):
                print(f'Removing existing dir {datadir}')
                shutil.rmtree(datadir)
        for subdir, flist in info.items():
            for url, md5 in flist:
                fpath, download = get_file(url, datadir=datadir, file_hash=md5, force_download=force_download)
                if download:
                    extract_archive(fpath, path=os.path.join(datadir, subdir))

        datapath = f'DATADIR_{dataset}={datadir}'
        with open(envfile) as f:
            lines = f.readlines()
        with open(envfile, 'w') as f:
            for l in lines:
                if f'DATADIR_{dataset}' in l:
                    l = f'export {datapath}\n'
                f.write(l)
        print(f'Updated dataset path in {envfile} to "{datapath}".')


if __name__ == '__main__':
    haid = DatasetsHub()
    print(haid.urls)

    info = haid.list()
    print(info)

    # haid.download_datsets('FINet')
    haid.download(name = 'SFID:latest')





from dataclasses import dataclass, field
from typing import Dict, Union, Literal


@dataclass
class AgentInfo:
    id: str = field(default=None, metadata={"help": "Agent's ID"})
    object: str = field(default="agent", metadata={"help": "Object type"})
    owner: str = field(default=None, metadata={"help": "Owner of the agent"})
    description: str = field(default=None, metadata={"help": "Description of the agent"})
    version: str = field(default="1.0", metadata={"help": "Version of the agent"})


import torch
from unet import UNet as _UNet

def unet_carvana(pretrained=False, scale=0.5):
    """
    UNet model trained on the Carvana dataset ( https://www.kaggle.com/c/carvana-image-masking-challenge/data ).
    Set the scale to 0.5 (50%) when predicting.
    """
    net = _UNet(n_channels=3, n_classes=2, bilinear=False)
    if pretrained:
        if scale == 0.5:
            checkpoint = 'https://github.com/milesial/Pytorch-UNet/releases/download/v3.0/unet_carvana_scale0.5_epoch2.pth'
        elif scale == 1.0:
            checkpoint = 'https://github.com/milesial/Pytorch-UNet/releases/download/v3.0/unet_carvana_scale1.0_epoch2.pth'
        else:
            raise RuntimeError('Only 0.5 and 1.0 scales are available')

        net.load_state_dict(torch.hub.load_state_dict_from_url(checkpoint, progress=True))

    return net




# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from typing_extensions import Literal

from ..._models import BaseModel
from .assistant_tool import AssistantTool
from .assistant_response_format_option import AssistantResponseFormatOption

__all__ = ["Assistant", "ToolResources", "ToolResourcesCodeInterpreter", "ToolResourcesFileSearch"]


class ToolResourcesCodeInterpreter(BaseModel):
    file_ids: Optional[List[str]] = None
    """
    A list of [file](https://platform.openai.com/docs/api-reference/files) IDs made
    available to the `code_interpreter`` tool. There can be a maximum of 20 files
    associated with the tool.
    """


class ToolResourcesFileSearch(BaseModel):
    vector_store_ids: Optional[List[str]] = None
    """
    The ID of the
    [vector store](https://platform.openai.com/docs/api-reference/vector-stores/object)
    attached to this assistant. There can be a maximum of 1 vector store attached to
    the assistant.
    """


class ToolResources(BaseModel):
    code_interpreter: Optional[ToolResourcesCodeInterpreter] = None

    file_search: Optional[ToolResourcesFileSearch] = None


class Assistant(BaseModel):
    id: str
    """The identifier, which can be referenced in API endpoints."""

    created_at: int
    """The Unix timestamp (in seconds) for when the assistant was created."""

    description: Optional[str] = None
    """The description of the assistant. The maximum length is 512 characters."""

    instructions: Optional[str] = None
    """The system instructions that the assistant uses.

    The maximum length is 256,000 characters.
    """

    metadata: Optional[object] = None
    """Set of 16 key-value pairs that can be attached to an object.

    This can be useful for storing additional information about the object in a
    structured format. Keys can be a maximum of 64 characters long and values can be
    a maximum of 512 characters long.
    """

    model: str
    """ID of the model to use.

    You can use the
    [List models](https://platform.openai.com/docs/api-reference/models/list) API to
    see all of your available models, or see our
    [Model overview](https://platform.openai.com/docs/models) for descriptions of
    them.
    """

    name: Optional[str] = None
    """The name of the assistant. The maximum length is 256 characters."""

    object: Literal["assistant"]
    """The object type, which is always `assistant`."""

    tools: List[AssistantTool]
    """A list of tool enabled on the assistant.

    There can be a maximum of 128 tools per assistant. Tools can be of types
    `code_interpreter`, `file_search`, or `function`.
    """

    response_format: Optional[AssistantResponseFormatOption] = None
    """Specifies the format that the model must output.

    Compatible with [GPT-4o](https://platform.openai.com/docs/models#gpt-4o),
    [GPT-4 Turbo](https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4),
    and all GPT-3.5 Turbo models since `gpt-3.5-turbo-1106`.

    Setting to `{ "type": "json_schema", "json_schema": {...} }` enables Structured
    Outputs which ensures the model will match your supplied JSON schema. Learn more
    in the
    [Structured Outputs guide](https://platform.openai.com/docs/guides/structured-outputs).

    Setting to `{ "type": "json_object" }` enables JSON mode, which ensures the
    message the model generates is valid JSON.

    **Important:** when using JSON mode, you **must** also instruct the model to
    produce JSON yourself via a system or user message. Without this, the model may
    generate an unending stream of whitespace until the generation reaches the token
    limit, resulting in a long-running and seemingly "stuck" request. Also note that
    the message content may be partially cut off if `finish_reason="length"`, which
    indicates the generation exceeded `max_tokens` or the conversation exceeded the
    max context length.
    """

    temperature: Optional[float] = None
    """What sampling temperature to use, between 0 and 2.

    Higher values like 0.8 will make the output more random, while lower values like
    0.2 will make it more focused and deterministic.
    """

    tool_resources: Optional[ToolResources] = None
    """A set of resources that are used by the assistant's tools.

    The resources are specific to the type of tool. For example, the
    `code_interpreter` tool requires a list of file IDs, while the `file_search`
    tool requires a list of vector store IDs.
    """

    top_p: Optional[float] = None
    """
    An alternative to sampling with temperature, called nucleus sampling, where the
    model considers the results of the tokens with top_p probability mass. So 0.1
    means only the tokens comprising the top 10% probability mass are considered.

    We generally recommend altering this or temperature but not both.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel
from .fine_tuning_job_wandb_integration import FineTuningJobWandbIntegration

__all__ = ["FineTuningJobWandbIntegrationObject"]


class FineTuningJobWandbIntegrationObject(BaseModel):
    type: Literal["wandb"]
    """The type of the integration being enabled for the fine-tuning job"""

    wandb: FineTuningJobWandbIntegration
    """The settings for your integration with Weights and Biases.

    This payload specifies the project that metrics will be sent to. Optionally, you
    can set an explicit display name for your run, add tags to your run, and set a
    default entity (team, username, etc) to be associated with your run.
    """



__appname__ = 'hepai'
__version__ = '1.2.0'
__version_suffix__ = 'Beta'  # version suffix, such as: 1.0.0-Dev, Beta, Stable
__author__ = 'ZhengDe Zhang'
__email__ = 'zdzhang@ihep.ac.cn, xionddb@ihep.ac.cn'
__backend__ = 'local'
__affiliation__ = 'CC, IHEP, CAS'
__supervisor__ = 'FaZhi Qi'
__collaborator__ = ''
__url__ = 'https://code.ihep.ac.cn/zdzhang/hai.git'




from typing import Dict, Any
import io
from ..._types import Stream
from .._resource import SyncAPIResource, AsyncAPIResource

from ..._return_class import (
    HWorkerListPage,
)
from ..._related_class import (
    WorkerInfo, HRemoteModel, WorkerStoppedInfo, WorkerStatusInfo,
)

# from ...openai_api._base_client import make_request_options
from openai._base_client import make_request_options


class Worker(SyncAPIResource):

    @property
    def prefix(self) -> str:
        return "/worker"
    
    def list_workers(self):
        return self._get(
            f"{self.prefix}/list_workers",
            cast_to=HWorkerListPage,
        )

    def get_info(
            self,
            worker_id: str = None,
            model_name: str = None,
            ) -> WorkerInfo:
        assert worker_id or model_name, "Either worker_id or model_name should be provided."
        payload = {
            "worker_id": worker_id,
            "model_name": model_name,
        }
        return self._post(
            f"{self.prefix}/get_worker_info",
            cast_to=WorkerInfo,
            body=payload,
        )

    def get_status(
            self,
            worker_id: str,
            refresh: bool = False,
            ) -> WorkerStatusInfo:
        payload = {
            "worker_id": worker_id,
            "refresh": refresh,
        }
        return self._post(
            f"{self.prefix}/get_worker_status",
            cast_to=WorkerStatusInfo,
            body=payload,
        )
    
    def stop(
            self,
            worker_id: str,
            permanent: bool = False,
            ) -> WorkerStoppedInfo:
        payload = {
            "worker_id": worker_id,
            "permanent": permanent,
        }
        return self._post(
            f"{self.prefix}/stop_worker",
            cast_to=WorkerStoppedInfo,
            body=payload,
        )
    
    def refresh_all(self):
        return self._get(
            f"{self.prefix}/refresh_all_workers",
            cast_to=Dict[str, Any],
        )
    
    def get_remote_model(
            self,
            worker_id: str = None,
            model_name: str = None,
            ) -> HRemoteModel:
        worker_info: WorkerInfo = self.get_info(worker_id=worker_id, model_name=model_name)
        if not isinstance(worker_info, WorkerInfo):
            raise ValueError(f"Failed to get remote model: {worker_info}")
        from ..._remote_model import LRemoteModel
        return LRemoteModel(name=model_name, worker_info=worker_info, worker_resource=self)

    def request(
            self,
            target: dict,  # 请求目标模型和函数, such as {"model": model_name, "function": "__call__"}
            args: list = None, 
            kwargs: dict = None,
        ):
        # get model and function
        model = target.get("model")
        function = target.get("function")  
        # check if need stream
        stream = kwargs.get("stream", False) if kwargs is not None else False

        from typing import cast, Mapping
        from openai._utils import deepcopy_minimal, extract_files      
        
        # set payload
        payload = dict()
        if args:
            payload["args"] = args
        if kwargs:
            payload["kwargs"] = kwargs
        
        # 适配文件上传，可从file字段中提取文件
        files = extract_files(cast(Mapping[str, object], kwargs), paths=[["file"]])
        payload = deepcopy_minimal(payload)
        
        # if True:
        #     payload = {'function_params': payload}
        
        extra_headers = None  # Headers
        extra_query = None  # Query
        extra_body = None  # Body
        # timeout = : float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        
        if files:
            extra_headers = {"Content-Type": "multipart/form-data", **(extra_headers or {})}

        if stream:
            return self._post(
                f"{self.prefix}/unified_gate/?model={model}&function={function}",
                body=payload,
                stream=True,
                stream_cls=Stream[Any],
                cast_to=Any,
                files=files,
                options=make_request_options(
                    extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body
            ),
            )
        return self._post(
            f"{self.prefix}/unified_gate/?model={model}&function={function}",
            cast_to=Any,
            body=payload,
            files=files,
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body
            ),
        )
    
    def register(
            self,
            model: HRemoteModel = None,
            daemon: bool = False,
            standalone: bool = False,
            )  -> WorkerInfo:
        
        from hepai import HWorkerAPP
        return HWorkerAPP.register_worker(model=model, daemon=daemon, standalone=standalone)
    

class AsyncWorker(AsyncAPIResource):
    """
    Asynchronous version of the Worker resource.
    """

    @property
    def prefix(self) -> str:
        return "/worker"

    async def list_workers(self) -> HWorkerListPage:
        """
        List all workers asynchronously.
        """
        return await self._get(
            f"{self.prefix}/list_workers",
            cast_to=HWorkerListPage,
        )

    async def get_info(
        self,
        worker_id: str = None,
        model_name: str = None,
    ) -> WorkerInfo:
        """
        Get worker info by `worker_id` or `model_name` asynchronously.
        Either worker_id or model_name should be provided.
        """
        assert worker_id or model_name, "Either worker_id or model_name should be provided."
        payload = {
            "worker_id": worker_id,
            "model_name": model_name,
        }
        return await self._post(
            f"{self.prefix}/get_worker_info",
            cast_to=WorkerInfo,
            body=payload,
        )

    async def get_status(
        self,
        worker_id: str,
        refresh: bool = False,
    ) -> WorkerStatusInfo:
        """
        Get worker status by `worker_id` asynchronously.
        Set `refresh` to True to refresh the worker status.
        """
        payload = {
            "worker_id": worker_id,
            "refresh": refresh,
        }
        return await self._post(
            f"{self.prefix}/get_worker_status",
            cast_to=WorkerStatusInfo,
            body=payload,
        )

    async def stop(
        self,
        worker_id: str,
        permanent: bool = False,
    ) -> WorkerStoppedInfo:
        """
        Stop a worker by `worker_id` asynchronously.
        If `permanent` is True, stop the worker permanently.
        """
        payload = {
            "worker_id": worker_id,
            "permanent": permanent,
        }
        return await self._post(
            f"{self.prefix}/stop_worker",
            cast_to=WorkerStoppedInfo,
            body=payload,
        )

    async def refresh_all(self) -> Dict[str, Any]:
        """
        Refresh all workers asynchronously.
        """
        return await self._get(
            f"{self.prefix}/refresh_all_workers",
            cast_to=Dict[str, Any],
        )

    async def get_remote_model(
        self,
        worker_id: str = None,
        model_name: str = None,
    ) -> HRemoteModel:
        """
        Get a remote model by `worker_id` or `model_name` asynchronously.
        """
        worker_info: WorkerInfo = await self.get_info(
            worker_id=worker_id,
            model_name=model_name,
        )
        if not isinstance(worker_info, WorkerInfo):
            raise ValueError(f"Failed to get remote model: {worker_info}")
        from ..._remote_model import LRemoteModel
        return LRemoteModel(name=model_name, worker_info=worker_info, worker_resource=self)

    async def request(
        self,
        target: dict,
        args: list = None,
        kwargs: dict = None,
    ):
        """
        Request the target worker model and function asynchronously.
        `target` is a dict, for example: {"model": "HRemoteModel", "function": "hello_world"}.
        """
        model = target.get("model")
        function = target.get("function")
        stream = kwargs.get("stream", False) if kwargs else False
        payload = {}

        if args:
            payload["args"] = args
        if kwargs:
            payload["kwargs"] = kwargs

        if stream:
            return await self._post(
                f"{self.prefix}/unified_gate/?model={model}&function={function}",
                body=payload,
                stream=True,
                stream_cls=Stream[Any],
                cast_to=Any,
            )
        return await self._post(
            f"{self.prefix}/unified_gate/?model={model}&function={function}",
            cast_to=Any,
            body=payload,
        )

    async def register(
        self,
        model: HRemoteModel = None,
        daemon: bool = False,
        standalone: bool = False,
    ) -> WorkerInfo:
        """
        Register a worker to the server asynchronously.
        """
        from hepai import HWorkerAPP
        # If HWorkerAPP.register_worker is purely synchronous,
        # we just call it directly. If there's an async method, use await accordingly.
        return HWorkerAPP.register_worker(model=model, daemon=daemon, standalone=standalone)


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List
from typing_extensions import Literal, Required, TypedDict

from .run_step_include import RunStepInclude

__all__ = ["StepListParams"]


class StepListParams(TypedDict, total=False):
    thread_id: Required[str]

    after: str
    """A cursor for use in pagination.

    `after` is an object ID that defines your place in the list. For instance, if
    you make a list request and receive 100 objects, ending with obj_foo, your
    subsequent call can include after=obj_foo in order to fetch the next page of the
    list.
    """

    before: str
    """A cursor for use in pagination.

    `before` is an object ID that defines your place in the list. For instance, if
    you make a list request and receive 100 objects, starting with obj_foo, your
    subsequent call can include before=obj_foo in order to fetch the previous page
    of the list.
    """

    include: List[RunStepInclude]
    """A list of additional fields to include in the response.

    Currently the only supported value is
    `step_details.tool_calls[*].file_search.results[*].content` to fetch the file
    search result content.

    See the
    [file search tool documentation](https://platform.openai.com/docs/assistants/tools/file-search#customizing-file-search-settings)
    for more information.
    """

    limit: int
    """A limit on the number of objects to be returned.

    Limit can range between 1 and 100, and the default is 20.
    """

    order: Literal["asc", "desc"]
    """Sort order by the `created_at` timestamp of the objects.

    `asc` for ascending order and `desc` for descending order.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

__all__ = ["AutoFileChunkingStrategyParam"]


class AutoFileChunkingStrategyParam(TypedDict, total=False):
    type: Required[Literal["auto"]]
    """Always `auto`."""



# from distutils.log import info
import os
import numpy as np
from PIL import Image
import torch
from torch.utils.data import DataLoader, random_split

import hai
from hai import AbstractModule, AbstractInput, AbstractOutput, AbstractQue
from hai import MODULES, SCRIPTS, IOS, Config

src_path = 'hai/modules/segmentation/Pytorch_UNet'
cwd = os.getcwd()
# from hai.modules.segmentation.Pytorch_UNet import train_hai
from . import train_api
from . import Pytorch_UNet
from . import predict_api
from . import evaluate_api
from . import argparse_config
from . import CarvanaDataset


@MODULES.register()
class UNet(AbstractModule):
    name = 'UNet'
    description = '2015 Classic Segmentation Algorithm UNet'
    

    def __init__(self):
        super().__init__()
        # self.default_cfg = f"{hai.root_path}/hai/configs/UNet/UNet_default_cfg.py"
        # self.default_cfg = train.get_args()
        self.default_cfg = argparse_config.get_args()
        # self.default_cfg.exporter = 'ImagesExporter'

    def __call__(self, img, *args, **kwargs):
        ## 输入的x是ndarray，需要前处理
        if isinstance(img, np.ndarray):
            img = Image.fromarray(np.uint8(img))

        ret = predict_api.predict_img(
                net=self.model,
                full_img=img,
                device=self.device,
                scale_factor=self.cfg.scale,
                out_threshold=self.cfg.mask_threshold,
        )
        return ret

    def build_model(self, *args, **kwargs):
        device = self.device
        cfg = self.cfg
        model = Pytorch_UNet(n_channels=3, n_classes=cfg.classes, bilinear=cfg.bilinear)
        if self.cfg.weights:
            model.load_state_dict(torch.load(self.cfg.weights, map_location=device))
        model.to(device=device)
        return model

    def train(self, cfg=None, *args, **kwargs):
        cfg = cfg if cfg else self.cfg
        train_api.run(args=cfg)
        # print('train') 
        # print(self.config)

    def predict(self, cfg=None, *args, **kwargs):
        cfg = cfg if cfg else self.cfg
        predict_api.run(args=cfg)

    def evaluate(self, cfg=None, *args, **kwargs):
        cfg = cfg if cfg else self.cfg
        # evaluate_api.run(args=cfg)
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        model = self.build_model()
        model.to(device=device)

        img_dir = f'{cfg.source}/imgs'
        mask_dir = f'{cfg.source}/masks'
        
        val_set = CarvanaDataset(img_dir, mask_dir, scale=cfg.scale)
        # train_set, val_set = random_split(val_set, [5000, 88], generator=torch.Generator().manual_seed(0))
        val_loader = DataLoader(val_set, 
            shuffle=False, drop_last=True, 
            batch_size=cfg.batch_size, num_workers=4, pin_memory=True)
        ret = evaluate_api.evaluate(
            net=model,
            dataloader=val_loader,
            device=device,
        )
        score = ret.detach().cpu().numpy()
        info = 'Score: {:.8f}'.format(score)
        # print(f'Evaluate score: {score}')
        return info


    



import glob
import os
import random
import shutil
import time
from pathlib import Path
from threading import Thread

import cv2
import math
import numpy as np
from PIL import Image, ExifTags
from sympy import im


from tqdm import tqdm

import damei as dm
import torch
from torch.utils.data import Dataset

from .datasets import _RepeatSampler, get_hash, img_formats, help_url, create_folder, load_image, exif_size
from .datasets import load_mosaic, letterbox, random_perspective, augment_hsv


class InfiniteDataLoader(torch.utils.data.dataloader.DataLoader):
    """ Dataloader that reuses workers.

    Uses same syntax as vanilla DataLoader.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        object.__setattr__(self, 'batch_sampler', _RepeatSampler(self.batch_sampler))
        self.iterator = super().__iter__()

    def __len__(self):
        return len(self.batch_sampler.sampler)

    def __iter__(self):
        for i in range(len(self)):
            yield next(self.iterator)



class LoadImagesAndLabels(Dataset):  # for training/testing
    def __init__(self, path, img_size=640, batch_size=16, augment=False, hyp=None, rect=False, image_weights=False,
                 cache_images=False, single_cls=False, stride=32, pad=0.0, rank=-1):
        self.img_size = img_size
        self.augment = augment
        self.hyp = hyp
        self.image_weights = image_weights
        self.rect = False if image_weights else rect
        self.mosaic = self.augment and not self.rect  # load 4 images at a time into a mosaic (only during training)
        self.mosaic_border = [-img_size // 2, -img_size // 2]
        self.stride = stride

        def img2label_paths(img_paths):
            # Define label paths as a function of image paths
            sa, sb = os.sep + 'images' + os.sep, os.sep + 'labels' + os.sep  # /images/, /labels/ substrings
            return [x.replace(sa, sb, 1).replace(os.path.splitext(x)[-1], '.txt') for x in img_paths]

        try:
            f = []  # image files
            for p in path if isinstance(path, list) else [path]:
                p = str(Path(p))  # os-agnostic
                parent = str(Path(p).parent) + os.sep
                if os.path.isfile(p):  # file
                    with open(p, 'r') as t:
                        t = t.read().splitlines()
                        f += [x.replace('./', parent) if x.startswith('./') else x for x in t]  # local to global path
                elif os.path.isdir(p):  # folder
                    f += glob.iglob(p + os.sep + '*.*')
                else:
                    raise Exception('%s does not exist' % p)
            self.img_files = sorted(
                [x.replace('/', os.sep) for x in f if os.path.splitext(x)[-1].lower() in img_formats])
            assert len(self.img_files) > 0, 'No images found'
        except Exception as e:
            raise Exception('Error loading data from %s: %s\nSee %s' % (path, e, help_url))

        # Check cache
        self.label_files = img2label_paths(self.img_files)  # labels
        cache_path = str(Path(self.label_files[0]).parent) + '.cache'  # cached labels
        if os.path.isfile(cache_path):
            cache = torch.load(cache_path)  # load
            if cache['hash'] != get_hash(self.label_files + self.img_files):  # dataset changed
                cache = self.cache_labels(cache_path)  # re-cache
        else:
            cache = self.cache_labels(cache_path)  # cache

        # Read cache
        cache.pop('hash')  # remove hash
        labels, shapes = zip(*cache.values())
        self.labels = list(labels)
        self.shapes = np.array(shapes, dtype=np.float64)
        self.img_files = list(cache.keys())  # update
        self.label_files = img2label_paths(cache.keys())  # update

        n = len(shapes)  # number of images
        bi = np.floor(np.arange(n) / batch_size).astype(np.int)  # batch index
        nb = bi[-1] + 1  # number of batches
        self.batch = bi  # batch index of image
        self.n = n

        # Rectangular Training
        if self.rect:
            # Sort by aspect ratio
            s = self.shapes  # wh
            ar = s[:, 1] / s[:, 0]  # aspect ratio
            irect = ar.argsort()
            self.img_files = [self.img_files[i] for i in irect]
            self.label_files = [self.label_files[i] for i in irect]
            self.labels = [self.labels[i] for i in irect]
            self.shapes = s[irect]  # wh
            ar = ar[irect]

            # Set training image shapes
            shapes = [[1, 1]] * nb
            for i in range(nb):
                ari = ar[bi == i]
                mini, maxi = ari.min(), ari.max()
                if maxi < 1:
                    shapes[i] = [maxi, 1]
                elif mini > 1:
                    shapes[i] = [1, 1 / mini]

            self.batch_shapes = np.ceil(np.array(shapes) * img_size / stride + pad).astype(np.int) * stride

        # Check labels
        create_datasubset, extract_bounding_boxes, labels_loaded = False, False, False
        nm, nf, ne, ns, nd = 0, 0, 0, 0, 0  # number missing, found, empty, datasubset, duplicate
        pbar = enumerate(self.label_files)
        if rank in [-1, 0]:
            pbar = tqdm(pbar)
        for i, file in pbar:
            l = self.labels[i]  # label
            if l is not None and l.shape[0]:
                assert l.shape[1] == 5, '> 5 label columns: %s' % file
                assert (l >= 0).all(), 'negative labels: %s' % file
                assert (l[:, 1:] <= 1).all(), 'non-normalized or out of bounds coordinate labels: %s' % file
                if np.unique(l, axis=0).shape[0] < l.shape[0]:  # duplicate rows
                    nd += 1  # print('WARNING: duplicate rows in %s' % self.label_files[i])  # duplicate rows
                if single_cls:
                    l[:, 0] = 0  # force dataset into single-class mode
                self.labels[i] = l
                nf += 1  # file found

                # Create subdataset (a smaller dataset)
                if create_datasubset and ns < 1E4:
                    if ns == 0:
                        create_folder(path='./datasubset')
                        os.makedirs('./datasubset/images')
                    exclude_classes = 43
                    if exclude_classes not in l[:, 0]:
                        ns += 1
                        # shutil.copy(src=self.img_files[i], dst='./datasubset/images/')  # copy image
                        with open('./datasubset/images.txt', 'a') as f:
                            f.write(self.img_files[i] + '\n')

                # Extract object detection boxes for a second stage classifier
                if extract_bounding_boxes:
                    p = Path(self.img_files[i])
                    img = cv2.imread(str(p))
                    h, w = img.shape[:2]
                    for j, x in enumerate(l):
                        f = '%s%sclassifier%s%g_%g_%s' % (p.parent.parent, os.sep, os.sep, x[0], j, p.name)
                        if not os.path.exists(Path(f).parent):
                            os.makedirs(Path(f).parent)  # make new output folder

                        b = x[1:] * [w, h, w, h]  # box
                        b[2:] = b[2:].max()  # rectangle to square
                        b[2:] = b[2:] * 1.3 + 30  # pad
                        b = dm.general.xywh2xyxy(b.reshape(-1, 4)).ravel().astype(np.int)

                        b[[0, 2]] = np.clip(b[[0, 2]], 0, w)  # clip boxes outside of image
                        b[[1, 3]] = np.clip(b[[1, 3]], 0, h)
                        assert cv2.imwrite(f, img[b[1]:b[3], b[0]:b[2]]), 'Failure extracting classifier boxes'
            else:
                ne += 1  # print('empty labels for image %s' % self.img_files[i])  # file empty
                # os.system("rm '%s' '%s'" % (self.img_files[i], self.label_files[i]))  # remove

            if rank in [-1, 0]:
                pbar.desc = 'Scanning labels %s (%g found, %g missing, %g empty, %g duplicate, for %g images)' % (
                    cache_path, nf, nm, ne, nd, n)
        if nf == 0:
            s = 'WARNING: No labels found in %s. See %s' % (os.path.dirname(file) + os.sep, help_url)
            print(s)
            assert not augment, '%s. Can not train without labels.' % s

        # Cache images into memory for faster training (WARNING: large datasets may exceed system RAM)
        self.imgs = [None] * n
        if cache_images:
            gb = 0  # Gigabytes of cached images
            pbar = tqdm(range(len(self.img_files)), desc='Caching images')
            self.img_hw0, self.img_hw = [None] * n, [None] * n
            for i in pbar:  # max 10k images
                self.imgs[i], self.img_hw0[i], self.img_hw[i] = load_image(self, i)  # img, hw_original, hw_resized
                gb += self.imgs[i].nbytes
                pbar.desc = 'Caching images (%.1fGB)' % (gb / 1E9)

    def cache_labels(self, path='labels.cache'):
        # Cache dataset labels, check images and read shapes
        x = {}  # dict
        pbar = tqdm(zip(self.img_files, self.label_files), desc='Scanning images', total=len(self.img_files))
        for (img, label) in pbar:
            try:
                l = []
                im = Image.open(img)
                im.verify()  # PIL verify
                shape = exif_size(im)  # image size
                assert (shape[0] > 9) & (shape[1] > 9), 'image size <10 pixels'
                if os.path.isfile(label):
                    with open(label, 'r') as f:
                        l = np.array([x.split() for x in f.read().splitlines()], dtype=np.float32)  # labels
                if len(l) == 0:
                    l = np.zeros((0, 5), dtype=np.float32)
                x[img] = [l, shape]
            except Exception as e:
                print('WARNING: Ignoring corrupted image and/or label %s: %s' % (img, e))

        x['hash'] = get_hash(self.label_files + self.img_files)
        torch.save(x, path)  # save for next time
        return x

    def __len__(self):
        return len(self.img_files)

    # def __iter__(self):
    #     self.count = -1
    #     print('ran dataset iter')
    #     #self.shuffled_vector = np.random.permutation(self.nF) if self.augment else np.arange(self.nF)
    #     return self

    def __getitem__(self, index):
        if self.image_weights:
            index = self.indices[index]

        hyp = self.hyp
        mosaic = self.mosaic and random.random() < hyp['mosaic']
        if mosaic:
            # Load mosaic
            img, labels = load_mosaic(self, index)
            shapes = None

            # MixUp https://arxiv.org/pdf/1710.09412.pdf
            if random.random() < hyp['mixup']:
                img2, labels2 = load_mosaic(self, random.randint(0, len(self.labels) - 1))
                r = np.random.beta(8.0, 8.0)  # mixup ratio, alpha=beta=8.0
                img = (img * r + img2 * (1 - r)).astype(np.uint8)
                labels = np.concatenate((labels, labels2), 0)

        else:
            # Load image
            img, (h0, w0), (h, w) = load_image(self, index)

            # Letterbox
            shape = self.batch_shapes[self.batch[index]] if self.rect else self.img_size  # final letterboxed shape
            img, ratio, pad = letterbox(img, shape, auto=False, scaleup=self.augment)
            shapes = (h0, w0), ((h / h0, w / w0), pad)  # for COCO mAP rescaling

            # Load labels
            labels = []
            x = self.labels[index]
            if x.size > 0:
                # Normalized xywh to pixel xyxy format
                labels = x.copy()
                labels[:, 1] = ratio[0] * w * (x[:, 1] - x[:, 3] / 2) + pad[0]  # pad width
                labels[:, 2] = ratio[1] * h * (x[:, 2] - x[:, 4] / 2) + pad[1]  # pad height
                labels[:, 3] = ratio[0] * w * (x[:, 1] + x[:, 3] / 2) + pad[0]
                labels[:, 4] = ratio[1] * h * (x[:, 2] + x[:, 4] / 2) + pad[1]

        if self.augment:
            # Augment imagespace
            if not mosaic:
                img, labels = random_perspective(img, labels,
                                                 degrees=hyp['degrees'],
                                                 translate=hyp['translate'],
                                                 scale=hyp['scale'],
                                                 shear=hyp['shear'],
                                                 perspective=hyp['perspective'])

            # Augment colorspace
            augment_hsv(img, hgain=hyp['hsv_h'], sgain=hyp['hsv_s'], vgain=hyp['hsv_v'])

            # Apply cutouts
            # if random.random() < 0.9:
            #     labels = cutout(img, labels)

        nL = len(labels)  # number of labels
        if nL:
            labels[:, 1:5] = dm.general.xyxy2xywh(labels[:, 1:5])  # convert xyxy to xywh
            labels[:, [2, 4]] /= img.shape[0]  # normalized height 0-1
            labels[:, [1, 3]] /= img.shape[1]  # normalized width 0-1

        if self.augment:
            # flip up-down
            if random.random() < hyp['flipud']:
                img = np.flipud(img)
                if nL:
                    labels[:, 2] = 1 - labels[:, 2]

            # flip left-right
            if random.random() < hyp['fliplr']:
                img = np.fliplr(img)
                if nL:
                    labels[:, 1] = 1 - labels[:, 1]

        labels_out = torch.zeros((nL, 6))
        if nL:
            labels_out[:, 1:] = torch.from_numpy(labels)

        # Convert
        img = img[:, :, ::-1].transpose(2, 0, 1)  # BGR to RGB, to 3x416x416
        img = np.ascontiguousarray(img)

        return torch.from_numpy(img), labels_out, self.img_files[index], shapes

    @staticmethod
    def collate_fn(batch):
        img, label, path, shapes = zip(*batch)  # transposed
        for i, l in enumerate(label):
            l[:, 0] = i  # add target image index for build_targets()
        return torch.stack(img, 0), torch.cat(label, 0), path, shapes






from typing import Union, Dict
from ..._types import Stream
from .._resource import SyncAPIResource

from ..._return_class import HAPIKeyListPage, HUserListPage

from ..._related_class import (
    APIKeyInfo, APIKeyDeletedInfo, UserInfo, UserDeletedInfo
)


class User(SyncAPIResource):

    @property
    def prefix(self) -> str:
        return "/user"
    
    def list_users(self):
        return self._get(
            f"{self.prefix}/list_users",
            cast_to=HUserListPage,
        )
    
    def create_user(self, **kwargs):
        payload = kwargs
        return self._post(
            f'{self.prefix}/create_user',
            body=payload,
            cast_to=UserInfo,
        )

    def delete_user(self, user_id: str):
        payload = {
            "user_id": str(user_id),
        }
        return self._post(
            f'{self.prefix}/delete_user',
            body=payload,
            cast_to=UserDeletedInfo,
        )
    
    def auth_user(self, username: str, password: str):
        payload = {
            "username": username,
            "password": password,
        }
        return self._post(
            f'{self.prefix}/auth_user',
            body=payload,
            cast_to=UserInfo,
        )
    
    


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

__all__ = ["ChatCompletionContentPartRefusalParam"]


class ChatCompletionContentPartRefusalParam(TypedDict, total=False):
    refusal: Required[str]
    """The refusal message generated by the model."""

    type: Required[Literal["refusal"]]
    """The type of the content part."""



from typing import Literal
from typing_extensions import override


class NotGiven:
    """
    A sentinel singleton class used to distinguish omitted keyword arguments
    from those passed in with the value None (which may have different behavior).

    For example:

    ```py
    def get(timeout: Union[int, NotGiven, None] = NotGiven()) -> Response: ...


    get(timeout=1)  # 1s timeout
    get(timeout=None)  # No timeout
    get()  # Default timeout behavior, which may not be statically known at the method definition.
    ```
    """

    def __bool__(self) -> Literal[False]:
        return False

    @override
    def __repr__(self) -> str:
        return "NOT_GIVEN"


""" Parts of the U-Net model """

import torch
import torch.nn as nn
import torch.nn.functional as F


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, bilinear=True):
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2])
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)



#! /usr/bin/env python
# coding=utf8

import grpc
import hello_pb2_grpc, hello_pb2


def run():
    '''
    模拟请求服务方法信息
    :return:
    '''
    conn = grpc.insecure_channel('localhost:50052')
    client = hello_pb2_grpc.GrpcServiceStub(channel=conn)
    skill = hello_pb2.Skill(name="engineer")
    request = hello_pb2.HelloRequest(data="xiao gang", skill=skill)
    respnse = client.hello(request)
    print("received:", respnse.result, respnse.map_result)


if __name__ == '__main__':
    run()



"""
seyolov5模型的配置文件
"""

# 用来加载模型的
model = dict(  # 加载模型时的参数
    type='SEYOLOv5',
    device='0',  # GPU
    weights='~/weights/xsensing/yolov5x_3.0.pt',  # ~代表用户根目录
    half=True,  # 半精度浮点数
    augment=False,
    use_se=False,
    backnone=None,
    neck=None,
    head=None,
    names=[  # 所有类别
        'person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train', 'truck', 'boat', 'traffic light',
        'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse', 'sheep', 'cow',
        'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag', 'tie', 'suitcase', 'frisbee',
        'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat', 'baseball glove', 'skateboard', 'surfboard',
        'tennis racket', 'bottle',
        'wine glass', 'cup', 'fork', 'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich', 'orange',
        'broccoli', 'carrot', 'hot dog', 'pizza', 'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed',
        'dining table', 'toilet', 'tv', 'laptop', 'mouse', 'remote', 'keyboard', 'cell phone', 'microwave', 'oven',
        'toaster', 'sink', 'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear',
        'hair drier', 'toothbrush'],
)

input_stream = dict(
    type='visible_input',
    enabled=False,
    source='~/datasets/xsensing/mm_data/slice_P4_1217/vis',  # 可以是文件夹、图像、list、rtsp
    img_size=640,  # 到这里值是定义了一个输入的流的源和resize的过程，加载到的是ndarray的数据，未归一化
)

post_process = dict(
    type='yolov5 post process',
    nms=dict(  # non max suppression
        conf_thres=0.25,  # 置信度阈值
        iou_thres=0.45,  # iou阈值
        agnostic_nms=False,
    ),
    filt_classes=dict(  # 仅指定的类别有效，不指定为None或为空列表
        enabled=True,
        valid_classes=[
            'person', 'bicycle', 'car', 'motorcycle', 'bus', 'truck', 'chair', 'bed'],
    ),
)

output_stream = dict(
    type='uaii_yolov5_output',
    save_dir=None,  # 如果有路径，保存为检测后的图像为{save_dir}/{stem}.jpg
    save_txt=False,  # 如果为True且有保存路径，保存检测结果的txt到{save_dir}/{stem}.txt
    save_conf=False,  # 如果为True，.txt内除了类别和bbox，还保存置信度
    print_ret=False,  # 打印结果到控制台
    que=dict(
        enabled=True,
        maxlen=5,
        wait=True,
    ),  # 如果为True，检测结果push到队列里。
)

# 训练配置
train = dict(
    optimizer=dict(
        type='Adam',
    ),
    max_epoches=100,
    augment=True,

)



import sys


def synchronized_with_attr(attr_name):
    def decorator(func):
        def synced_func(*args, **kws):
            self = args[0]
            lock = getattr(self, attr_name)
            with lock:
                return func(*args, **kws)

        return synced_func

    return decorator


def truncate(ori_str, length=100):
    if not ori_str:
        return ""
    return ori_str[:length] + "..." if len(ori_str) > length else ori_str


def python_version_bellow(version):
    if not version:
        return False

    sp = [int(s) for s in version.split(".")]
    for i in range(len(sp) if len(sp) <= 3 else 3):
        if sp[i] == sys.version_info[i]:
            continue

        return sys.version_info[i] < sp[i]

    return False


import imp


import os, sys

model_dir = 'hai/modules/segmentation/Pytorch_UNet'
# train
from hai.modules.segmentation.Pytorch_UNet.utils.data_loading import BasicDataset, CarvanaDataset
from hai.modules.segmentation.Pytorch_UNet.utils.dice_score import dice_loss
from hai.modules.segmentation.Pytorch_UNet.evaluate import evaluate
from hai.modules.segmentation.Pytorch_UNet.unet import UNet as Pytorch_UNet
# predict
from hai.modules.segmentation.Pytorch_UNet.utils.utils import plot_img_and_mask
# evaluate
from hai.modules.segmentation.Pytorch_UNet.utils.dice_score import multiclass_dice_coeff, dice_coeff


from .unet_api import *






import os, sys
from pathlib import Path    
here = Path(__file__).parent
from ..version import __appname__

# /home/xxx/VSProjects/hai
ROOT_DIR = here.parent.parent
ROOT_PATH = ROOT_DIR

# /home/xxx/.hepai/weights
WEIGHTS_ROOT = f'{Path.home()}/.{__appname__}/weights'

# /home/xxx/.hepai/datasets
DATASETS_ROOT = f'{Path.home()}/.{__appname__}/datasets'

# hai_api
API_FOLD_NAME = f'{__appname__}_api'



# 可以是文件夹、图像、list、rtsp
source = '~/datasets/xsensing/mm_data/slice_P4_1217/vis'
# 到这里值是定义了一个输入的流的源和resize的过程，加载到的是ndarray的数据，未归一化
img_size = 640

input_stream = dict(
    type='visible_dataloader_input_stream',
    enabled=False,
    que=dict(),  # 输入只需要知道队列名字就可以
)

output_stream = dict(
    type='visible_dataloader_output_stream',
    enabled=True,
    que=dict(
        enabled=True,
        wait=False,
    ),
)



import hai
from hai import AbstractModule, AbstractInput, AbstractOutput, AbstractQue
from hai import MODULES, SCRIPTS, IOS, Config


@hai.MODULES.register_module(name='ResNet')
class ResNet(AbstractModule):
    name = 'ResNet'
    description = '2015 Classic 残差连接网络ResNet18 34 50 101 152 for classification'

    def __init__(self, config):
        super().__init__(config)
        self.config = config


from dataclasses import dataclass, field, asdict
from typing import Optional, List, Union, Dict
from datetime import datetime
import uuid


@dataclass
class UserGroupInfo:
    name: str = field(default=None, metadata={'description': 'Name of the user group.'})
    id: Optional[str] = field(default_factory=str, metadata={'description': 'Unique identifier for the user group, typically a UUID.'})
    description: Optional[str] = field(default=None, metadata={'description': 'Description of the user group.'})
    create_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the user group was created.'})
    created_by: Optional[str] = field(default=None, metadata={'description': 'Identifier for the creator of the user group record.'})
    update_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the user group record was last updated.'})
    updated_by: Optional[str] = field(default=None, metadata={'description': 'Identifier for who last updated the user group record.'})
    remarks: Optional[str] = field(default=None, metadata={'description': 'Additional remarks or notes about the user group.'})
    del_flag: bool = field(default=False, metadata={'description': 'Flag indicating if the user group is deleted.'})

    def to_dict(self) -> dict:
        return asdict(self)
    

    
@dataclass
class UserLevelInfo:
    user_level: str = field(default='free', metadata={'description': 'User level or role, e.g., admin, user.'})
    id: Optional[int] = field(default=None, metadata={'description': 'Unique identifier for the user level, typically an integer.'})
    expiration_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the user level expires.'})
    create_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the user level was created.'})
    created_by: Optional[str] = field(default=None, metadata={'description': 'Identifier for the creator of the user level record.'})
    update_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the user level record was last updated.'})
    updated_by: Optional[str] = field(default=None, metadata={'description': 'Identifier for who last updated the user level record.'})
    remarks: Optional[str] = field(default=None, metadata={'description': 'Additional remarks or notes about the user level.'})
    del_flag: bool = field(default=False, metadata={'description': 'Flag indicating if the user level is deleted.'})

    def __post_init__(self):
        # 如果没有传入expiration_time，自动设置PLUS用户的过期时间为1个月
        if self.expiration_time is None and self.user_level == 'plus':
            from dateutil.relativedelta import relativedelta
            self.expiration_time = datetime.now() + relativedelta(months=1)
        pass

    @classmethod
    def allowed_user_levels(cls) -> List[str]:
        return ['public', 'free', 'member', 'plus', 'app_admin', 'admin']

    def to_dict(self) -> dict:
        return asdict(self)
    
    def str2int(self) -> dict:
        mapping = {k: v for v, k in enumerate(self.allowed_user_levels())}
        return mapping.get(self.user_level, 0)
    
    @classmethod
    def get_mapping(cls) -> dict:
        return {k: v for v, k in enumerate(cls.allowed_user_levels())}
    
    def to_str(self) -> str:
        return self.user_level.lower()
    
    def str(self) -> str:
        return self.user_level.lower()
    

    @classmethod
    def int2str(cls, user_level: int = 1) -> str:
        allowed_user_levels = UserLevelInfo.allowed_user_levels()
        assert user_level in range(len(allowed_user_levels)), f"Invalid user_level: {user_level}"
        return allowed_user_levels[user_level]
    
    def __str__(self) -> str:
        return str(self.user_level)
    
    def __int__(self) -> int:
        return self.str2int()
    
    def __repr__(self) -> str:
        return f"UserLevelInfo(user_level={self.user_level!r}, id={self.id!r}, expiration_time={self.expiration_time!r}."


@dataclass
class UserInfo:
    id: str = field(default_factory=str, metadata={'description': 'Unique identifier for the user, typically a UUID.'})
    username: Optional[str] = field(default=None, metadata={'description': 'Username of the user.'})
    password: Optional[str] = field(default=None, metadata={'description': 'Password of the user, typically hashed.'})
    user_level: Optional[UserLevelInfo] = field(default=None, metadata={'description': 'User level or role, e.g., admin, user.'})
    user_groups: Optional[List[UserGroupInfo]] = field(default=None, metadata={'description': 'List of user groups that the user belongs to.'})
    phone: Optional[str] = field(default=None, metadata={'description': 'Phone number of the user.'})
    auth_type: Optional[str] = field(default="local", metadata={'description': 'Authentication type for the user, e.g., OAuth, SSO.'})
    create_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the user was created.'})
    created_by: Optional[str] = field(default=None, metadata={'description': 'Identifier for the creator of the user record.'})
    update_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the user record was last updated.'})
    updated_by: Optional[str] = field(default=None, metadata={'description': 'Identifier for who last updated the user record.'})
    remarks: Optional[str] = field(default=None, metadata={'description': 'Additional remarks or notes about the user.'})
    del_flag: bool = field(default=False, metadata={'description': 'Flag indicating if the user is deleted.'})
    umt_id: Optional[int] = field(default=None, metadata={'description': 'Foreign key or reference ID for external systems.'})
    email: Optional[str] = field(default=None, metadata={'description': 'Email address of the user.'})
    cluster_uid: Optional[int] = field(default=None, metadata={'description': 'Cluster unique identifier, possibly used in multi-tenant systems.'})
    balance: Optional[float] = field(default=0.0, metadata={'description': 'User balance, possibly for billing or credits.'})

    def __post_init__(self):
        if isinstance(self.id, uuid.UUID):  # 如果id是uuid4类型，转换为str
            self.id = str(self.id)
        # 允许传入user_level为None或str, None时默认为free
        if self.user_level is None:
            self.user_level = 'free'
        if isinstance(self.user_level, str):
            if self.user_level not in UserLevelInfo.allowed_user_levels():
                raise ValueError(f"Invalid user_level: {self.user_level}")
            self.user_level = UserLevelInfo(user_level=self.user_level)
        elif isinstance(self.user_level, dict):
            self.user_level = UserLevelInfo(**self.user_level)
        # 如果未传入，自动设置group为default
        if self.user_groups is None:
            self.user_groups = [UserGroupInfo(name="default")]
        elif isinstance(self.user_groups, list):
            if isinstance(self.user_groups[0], dict):
                self.user_groups = [UserGroupInfo(**group) for group in self.user_groups]
                
            

    def __repr__(self) -> str:
        return f"UserInfo(id={self.id!r}, username={self.username!r}, phone={self.phone!r}, email={self.email!r}, del_flag={self.del_flag!r})"

    def to_dict(self) -> dict:
        return asdict(self)
    
    def group_ids(self) -> List[str]:
        return [group.id for group in self.user_groups]
    
    def group_names(self) -> List[str]:
        return [group.name for group in self.user_groups]
    

@dataclass
class UserDeletedInfo:
    id: str = field(default_factory=str, metadata={'description': 'Unique identifier for the user, typically a UUID.'})
    username: Optional[str] = field(default=None, metadata={'description': 'Username of the user.'})
    deleted: bool = field(default=False, metadata={'description': 'Flag indicating if the user is deleted.'})


@dataclass
class APIKeyInfo:
    id: str = field(default_factory=str, metadata={'description': 'Unique identifier for the API key, typically a UUID.'})
    api_key: str = field(default_factory=str, metadata={'description': 'Value of the API key.'})
    alias: Optional[str] = field(default=None, metadata={'description': 'Alias name for the API key.'})
    user_id: Optional[str] = field(default=None, metadata={'description': 'User ID that the API key belongs to.'})
    username: Optional[str] = field(default=None, metadata={'description': 'Username that the API key belongs to.'})
    expiration_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the API key expires.'})
    usage_description: Optional[str] = field(default=None, metadata={'description': 'Description of the API key usage.'})
    create_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the API key was created.'})
    created_by: Optional[str] = field(default=None, metadata={'description': 'Identifier for the creator of the API key record.'})
    update_time: Optional[datetime] = field(default=None, metadata={'description': 'Timestamp when the API key record was last updated.'})
    updated_by: Optional[str] = field(default=None, metadata={'description': 'Identifier for who last updated the API key record.'})
    allowed_models: Optional[Union[str, Dict]] = field(default="all", metadata={'description': 'Allowed models for the API key.'})
    remarks: Optional[str] = field(default=None, metadata={'description': 'Additional remarks or notes about the API key.'})
    app_group: Optional[str] = field(default=None, metadata={'description': 'App group that the API key belongs to.'})
    del_flag: bool = field(default=False, metadata={'description': 'Flag indicating if the API key is deleted.'})
    

    def __post_init__(self):
        if isinstance(self.id, uuid.UUID):
            self.id = str(self.id)

    
    def __repr__(self) -> str:
        return f"APIKeyInfo(id={self.id!r}, alias={self.alias!r}, api_key={self.api_key!r}, user_id={self.user_id!r}, expiration_time={self.expiration_time!r})"

    def to_str(self) -> str:
        return self.api_key

    def to_dict(self) -> dict:
        return asdict(self)
    
@dataclass
class APIKeyDeletedInfo:
    id: str = field(default_factory=str, metadata={'description': 'Unique identifier for the API key, typically a UUID.'})
    api_key: str = field(default_factory=str, metadata={'description': 'Value of the API key.'})
    deleted: bool = field(default=False, metadata={'description': 'Flag indicating if the API key is deleted.'})
    
    def __str__(self) -> str:
        return str(self.api_key)
    
    def __repr__(self) -> str:
        return f"APIKeyDeletedInfo(id={self.id!r}, api_key={self.api_key!r}, deleted={self.deleted!r})"
                
    
    def to_dict(self) -> dict:
        return asdict(self)


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.


from ..._models import BaseModel

__all__ = ["StaticFileChunkingStrategy"]


class StaticFileChunkingStrategy(BaseModel):
    chunk_overlap_tokens: int
    """The number of tokens that overlap between chunks. The default value is `400`.

    Note that the overlap must not exceed half of `max_chunk_size_tokens`.
    """

    max_chunk_size_tokens: int
    """The maximum number of tokens in each chunk.

    The default value is `800`. The minimum value is `100` and the maximum value is
    `4096`.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.


from .._models import BaseModel

__all__ = ["BatchRequestCounts"]


class BatchRequestCounts(BaseModel):
    completed: int
    """Number of requests that have been completed successfully."""

    failed: int
    """Number of requests that have failed."""

    total: int
    """Total number of requests in the batch."""





from typing import Union, Dict
from ...openai_api import Stream
from .._resource import SyncAPIResource

from ..._return_class import (
    HAPIKeyListPage,
)
from ..._related_class import (
    APIKeyInfo, APIKeyDeletedInfo,
)

class Key(SyncAPIResource):

    @property
    def prefix(self) -> str:
        return "/key"
    
    def list_api_keys(self):
        return self._get(
            f"{self.prefix}/list_api_keys",
            cast_to=HAPIKeyListPage,
        )
    
    def create_api_key(
            self,
            *,
            key_name: str = "Default",
            valid_time: int = 30,
            user_id: str = None,
            umt_id: str = None,
            allowed_models: Union[str, Dict] = "all",
            remarks: str = "",
            ):
        payload = {
            "key_name": key_name,
            "valid_time": valid_time,
            "user_id": user_id,
            "umt_id": umt_id,
            "allowed_models": allowed_models,
            "remarks": remarks,
        }
        return self._post(
            f"{self.prefix}/create_api_key",
            body=payload,
            cast_to=APIKeyInfo,
        )
    
    def delete_api_key(self, api_key_id: str):
        payload = {
            "api_key_id": api_key_id,
        }
        return self._post(
            f"{self.prefix}/delete_api_key",
            body=payload,
            cast_to=APIKeyDeletedInfo,
        )
    
    def get_info(self, api_key: str, version: str = "v2"):
        payload = {
            "api_key": api_key,
        }
        url = f"{self.prefix}/verify_api_key" if version == "v2" else f"/verify_api_key"
        return self._post(
            f"{url}",
            body=payload,
            cast_to=Dict,
        )
    
    def fetch_api_key(self, username: str):
        payload = {
            "username": username,
        }
        return self._post(
            f"{self.prefix}/fetch_api_key",
            body=payload,
            cast_to=APIKeyInfo,
        )
        



from typing import Dict, List, Literal, Optional, Union, Generator, Callable
from dataclasses import dataclass, field

from fastapi import FastAPI, APIRouter, Request, HTTPException
from fastapi import UploadFile, File, Form, Query, Body
from fastapi.background import BackgroundTasks
from fastapi.responses import JSONResponse, StreamingResponse, FileResponse, HTMLResponse

import asyncio
from asyncio import Semaphore
import threading
from pydantic import BaseModel
from logging import Logger
import httpx
import inspect
# import markdown

from ._worker_class import HWorkerConfig, HRemoteModel, CommonWorker, ModelResourceInfo
from ._related_class import WorkerStoppedInfo, WorkerInfoRequest, WorkerUnifiedGateRequest
from . import utils
from .singletons import authorizer

class FunctionParamsItem(BaseModel):
    args: List = []
    kwargs: Dict = {}


def get_fastapi_init_params():
    # 获取 FastAPI 的 __init__ 方法签名
    init_signature = inspect.signature(FastAPI.__init__)
    # 提取参数名（排除 'self' 和可变参数）
    params = []
    for name, param in init_signature.parameters.items():
        if name == "self":
            continue
        if param.kind in (
            inspect.Parameter.VAR_POSITIONAL,  # *args
            inspect.Parameter.VAR_KEYWORD       # **kwargs
        ):
            continue
        params.append(name)
    return params

class HWorkerAPP(FastAPI):
    """
    FastAPI app for worker
    多线程运行，支持多个模型
    """

    def __init__(
            self,
            models: HRemoteModel | List[HRemoteModel], 
            worker_config: HWorkerConfig = None,  # Alias of config
            logger: Logger = None,
            **worker_overrides,
            ):
        # 获取父类的所有可能参数并传递给父类
        # 获取 FastAPI 接受的参数名
        fastapi_params = get_fastapi_init_params()
        # 筛选出有效参数
        fastapi_kwargs = {
            k: v for k, v in worker_overrides.items()
            if k in fastapi_params
        }
        # 从 worker_overrides 中删除 FastAPI 的参数
        for k in fastapi_kwargs.keys():
            worker_overrides.pop(k)
        super().__init__(**fastapi_kwargs)
        self.logger = self.get_logger(logger)
        worker_config = worker_config if worker_config is not None else HWorkerConfig()
        assert isinstance(worker_config, HWorkerConfig), f"worker_config should be an instance of HWorkerConfig"
        worker_config.update_from_dict(worker_overrides)
        
        # 用于控制模型访问的信号量
        self.limit_model_concurrency = worker_config.limit_model_concurrency
        self.model_semaphore: Semaphore = asyncio.Semaphore(self.limit_model_concurrency)  # 初始化信号量
        self.global_counter = 0
        if worker_config.enable_secret_key:
            seed = utils.get_simple_machine_seed()
            worker_secret_key = utils.gen_one_key(prefix='sk-', lenth=47, seed=seed)
            self.logger.info(f"Worker secret key: `{worker_secret_key}`, please pass it in the `Authorization` header when you call this worker.")
            authorizer.secret_key = worker_secret_key
        else:
            worker_secret_key = None
        self.worker_secret_key = worker_secret_key

        self.worker = CommonWorker(
            app=self, models=models, worker_config=worker_config, 
            logger=self.logger)
        self._init_routers(config=worker_config)

        self._docs_content_cache = None  # 缓存docs内容
        self._html_mk_content_cache = None  # 缓存markdown内容

    def _init_routers(self, config: HWorkerConfig):
        
        # 1 index router
        index_router = APIRouter(prefix="", tags=["base"])
        index_router.get("/")(self.index)
        self.include_router(index_router)

        # 2 worker router
        worker_router = self.get_worker_router()
        self.include_router(worker_router)
        
        # 3 llm router
        if config.enable_llm_router:
            from .routers.llm_router import LLMRouterGroup
            llm_rg = LLMRouterGroup(prefix=config.route_prefix, parent_app=self)
            self.include_router(llm_rg.router, prefix=llm_rg.prefix, tags=llm_rg.tags)
        
    def get_worker_router(self):
        router_prefix = self.worker.config_dict.get("route_prefix", "/apiv2")
        router = APIRouter(prefix=router_prefix, tags=["worker"])
        router.post("/worker_unified_gate/")(self.worker_unified_gate)
        router.post("/worker_unified_gate/{function}")(self.worker_unified_gate)
        router.post("/worker_unified_gate/{model}/{function}")(self.worker_unified_gate)  # 多模型模式下，需要指定模型
        router.get("/worker_get_status")(self.worker_get_status)
        router.post("/shutdown_worker")(self.shutdown_worker)
        
        router.post("/worker/get_worker_info")(self.get_worker_info)  # 这个路由是为了与controller相同的格式，使得client也能调用
        router.post("/worker/unified_gate")(self.worker_unified_gate)  # 这个路由是为了与controller相同的格式，使得client也能调用
        router.get("/worker/models")(self.get_models)
        return router

    @classmethod
    def get_logger(cls, logger: Logger = None):
        if logger is None:
            try:
                from ...utils._logger import Logger
                logger = Logger.get_logger("worker_app.py")
            except:
                import logging
                logger = logging.getLogger("worker_app.py")
        return logger

    @property
    def host(self):
        return self.worker.network_info.host
    
    @property
    def port(self):
        return self.worker.network_info.port
    

    def get_queue_length(self):
        model_semaphore: asyncio.Semaphore = self.model_semaphore
        if model_semaphore is None:
            return 0
        try:
            _value = model_semaphore._value
            _num_waiters = len(model_semaphore._waiters) if hasattr(model_semaphore, "_waiters") and model_semaphore._waiters is not None else 0
            return self.limit_model_concurrency - _value + _num_waiters
        except Exception:
            return 0

    async def release_model_semaphore(self):
        if self.model_semaphore is not None:
            self.model_semaphore.release()

    async def index(self):
        worker_name = self.worker.worker_id
        fastapi_docs_url = f"http://{self.host}:{self.port}/docs"
        # 缓存docs内容，避免重复请求
        if self._docs_content_cache is None:
            async with httpx.AsyncClient() as client:
                response = await client.get(fastapi_docs_url)
                self._docs_content_cache = response.text
        docs_content = self._docs_content_cache

        model_names = [m.name for m in self.worker.models]
        markdown_content = f"""
# HepAI Worker API
- Worker Name: **{worker_name}**
- Models: **{model_names}**
"""
        mrs = self.worker.get_model_resource_info()
        for mr in mrs:
            markdown_content += f"""
### {mr.model_name}
- Type: {mr.model_type}
- Functions: {mr.model_functions}
"""
        # 缓存markdown转换结果
        if self._html_mk_content_cache is None:
            import markdown
            self._html_mk_content_cache = markdown.markdown(markdown_content)
        html_mk_content = self._html_mk_content_cache

        content = f"""
            <html>
                <head>
                    <title>HepAI Worker Info</title>
                </head>
                <body>
                    <div style="padding-left: 20px;">
                        {html_mk_content}
                    </div>
                    <div>
                        {docs_content}
                    </div>
                </body>
            </html>
            """
        return HTMLResponse(content=content)

    async def worker_unified_gate(
            self,
            function_params: FunctionParamsItem,
            model: str = None,
            function: str = "__call__",
            ):
        self.global_counter += 1
        await self.model_semaphore.acquire()
        try:
            rst = await self.worker.unified_gate_async(
                model=model,
                function=function, 
                args=function_params.args,
                kwargs=function_params.kwargs)
        except Exception as e:
            await self.release_model_semaphore()
            raise e
        background_tasks = BackgroundTasks()
        background_tasks.add_task(self.release_model_semaphore)
        return rst
    
    async def worker_get_status(self):
        return self.worker.get_status_info().to_dict()
    
    async def get_worker_info(
            self, 
            worker_info_request: WorkerInfoRequest,
            # user_auth: HAPIKeyAuth = api_key_auth,
            ) -> JSONResponse:
        """
        与controller的worker_info接口一致，以便client调用
        """
        return self.worker.get_worker_info()

    async def get_models(self):
        mrs: List[ModelResourceInfo] = self.worker.get_model_resource_info()
        
        return {
            "object": "list",
            "data": [mr.to_dict() for mr in mrs],
        }
    
    
    async def shutdown_worker(self, background_tasks: BackgroundTasks):
        """接收来自Ctrl的关闭worker的信息"""
        background_tasks.add_task(self.worker.shutdown_worker)
        self.worker._is_deleted_in_controller = True
        wid = self.worker.worker_id
        return WorkerStoppedInfo(
            id=wid, stopped=True, 
            message=f"Worker `{wid}` shutdown",
            shutdown=True,
            )

    @classmethod
    def register_worker(
            cls,
            model: HRemoteModel = None,
            worker_config: HWorkerConfig = None,
            daemon: bool = False,
            standalone: bool = False,
            **kwargs,
            ):
        """注册HModelWorker到HaiDDF"""
        from .utils import run_standlone_worker_demo
        if standalone:  # 独立程序模式，用户测试程序
            # cls().logger.info("Running `worker` in standalone mode.")
            print("Running `worker` in standalone mode.")
            return run_standlone_worker_demo()
        
        assert model is not None, f"Model should be not None"
        import uvicorn
        app: FastAPI = HWorkerAPP(model, worker_config=worker_config, **kwargs)
        def run_uvicron():
            uvicorn.run(app, host=app.host, port=app.port)

        if daemon:  # 守护进程模式，app workre在后台运行
            t = threading.Thread(target=run_uvicron, daemon=True)
            t.start()
            return app.worker.get_worker_info()
        else:  # 正常模式，app worker在前台运行
            run_uvicron()

    def run(self):
        import uvicorn
        uvicorn.run(self, host=self.host, port=self.port)










import argparse
import logging
import sys
from pathlib import Path

import torch
import torch.nn as nn
import torch.nn.functional as F
import wandb
from torch import optim
from torch.utils.data import DataLoader, random_split
from tqdm import tqdm

from .utils.data_loading import BasicDataset, CarvanaDataset
from .utils.dice_score import dice_loss
from .evaluate import evaluate
from .unet import UNet

dir_img = Path('./data/imgs/')
dir_mask = Path('./data/masks/')
dir_checkpoint = Path('./checkpoints/')


def train_net(net,
              device,
              epochs: int = 5,
              batch_size: int = 1,
              learning_rate: float = 1e-5,
              val_percent: float = 0.1,
              save_checkpoint: bool = True,
              img_scale: float = 0.5,
              amp: bool = False):
    # 1. Create dataset
    try:
        dataset = CarvanaDataset(dir_img, dir_mask, img_scale)
    except (AssertionError, RuntimeError):
        dataset = BasicDataset(dir_img, dir_mask, img_scale)

    # 2. Split into train / validation partitions
    n_val = int(len(dataset) * val_percent)
    n_train = len(dataset) - n_val
    train_set, val_set = random_split(dataset, [n_train, n_val], generator=torch.Generator().manual_seed(0))

    # 3. Create data loaders
    loader_args = dict(batch_size=batch_size, num_workers=4, pin_memory=True)
    train_loader = DataLoader(train_set, shuffle=True, **loader_args)
    val_loader = DataLoader(val_set, shuffle=False, drop_last=True, **loader_args)

    # (Initialize logging)
    experiment = wandb.init(project='U-Net', resume='allow', anonymous='must')
    experiment.config.update(dict(epochs=epochs, batch_size=batch_size, learning_rate=learning_rate,
                                  val_percent=val_percent, save_checkpoint=save_checkpoint, img_scale=img_scale,
                                  amp=amp))

    logging.info(f'''Starting training:
        Epochs:          {epochs}
        Batch size:      {batch_size}
        Learning rate:   {learning_rate}
        Training size:   {n_train}
        Validation size: {n_val}
        Checkpoints:     {save_checkpoint}
        Device:          {device.type}
        Images scaling:  {img_scale}
        Mixed Precision: {amp}
    ''')

    # 4. Set up the optimizer, the loss, the learning rate scheduler and the loss scaling for AMP
    optimizer = optim.RMSprop(net.parameters(), lr=learning_rate, weight_decay=1e-8, momentum=0.9)
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'max', patience=2)  # goal: maximize Dice score
    grad_scaler = torch.cuda.amp.GradScaler(enabled=amp)
    criterion = nn.CrossEntropyLoss()
    global_step = 0

    # 5. Begin training
    for epoch in range(1, epochs+1):
        net.train()
        epoch_loss = 0
        with tqdm(total=n_train, desc=f'Epoch {epoch}/{epochs}', unit='img') as pbar:
            for batch in train_loader:
                images = batch['image']
                true_masks = batch['mask']

                assert images.shape[1] == net.n_channels, \
                    f'Network has been defined with {net.n_channels} input channels, ' \
                    f'but loaded images have {images.shape[1]} channels. Please check that ' \
                    'the images are loaded correctly.'

                images = images.to(device=device, dtype=torch.float32)
                true_masks = true_masks.to(device=device, dtype=torch.long)

                with torch.cuda.amp.autocast(enabled=amp):
                    masks_pred = net(images)
                    loss = criterion(masks_pred, true_masks) \
                           + dice_loss(F.softmax(masks_pred, dim=1).float(),
                                       F.one_hot(true_masks, net.n_classes).permute(0, 3, 1, 2).float(),
                                       multiclass=True)

                optimizer.zero_grad(set_to_none=True)
                grad_scaler.scale(loss).backward()
                grad_scaler.step(optimizer)
                grad_scaler.update()

                pbar.update(images.shape[0])
                global_step += 1
                epoch_loss += loss.item()
                experiment.log({
                    'train loss': loss.item(),
                    'step': global_step,
                    'epoch': epoch
                })
                pbar.set_postfix(**{'loss (batch)': loss.item()})

                # Evaluation round
                division_step = (n_train // (10 * batch_size))
                if division_step > 0:
                    if global_step % division_step == 0:
                        histograms = {}
                        for tag, value in net.named_parameters():
                            tag = tag.replace('/', '.')
                            histograms['Weights/' + tag] = wandb.Histogram(value.data.cpu())
                            histograms['Gradients/' + tag] = wandb.Histogram(value.grad.data.cpu())

                        val_score = evaluate(net, val_loader, device)
                        scheduler.step(val_score)

                        logging.info('Validation Dice score: {}'.format(val_score))
                        experiment.log({
                            'learning rate': optimizer.param_groups[0]['lr'],
                            'validation Dice': val_score,
                            'images': wandb.Image(images[0].cpu()),
                            'masks': {
                                'true': wandb.Image(true_masks[0].float().cpu()),
                                'pred': wandb.Image(masks_pred.argmax(dim=1)[0].float().cpu()),
                            },
                            'step': global_step,
                            'epoch': epoch,
                            **histograms
                        })

        if save_checkpoint:
            Path(dir_checkpoint).mkdir(parents=True, exist_ok=True)
            torch.save(net.state_dict(), str(dir_checkpoint / 'checkpoint_epoch{}.pth'.format(epoch)))
            logging.info(f'Checkpoint {epoch} saved!')


def run(opt):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    logging.info(f'Using device {device}')

    # Change here to adapt to your data
    # n_channels=3 for RGB images
    # n_classes is the number of probabilities you want to get per pixel
    net = UNet(n_channels=3, n_classes=args.classes, bilinear=args.bilinear)

    logging.info(f'Network:\n'
                 f'\t{net.n_channels} input channels\n'
                 f'\t{net.n_classes} output channels (classes)\n'
                 f'\t{"Bilinear" if net.bilinear else "Transposed conv"} upscaling')

    if args.load:
        net.load_state_dict(torch.load(args.load, map_location=device))
        logging.info(f'Model loaded from {args.load}')

    net.to(device=device)
    try:
        train_net(net=net,
                  epochs=args.epochs,
                  batch_size=args.batch_size,
                  learning_rate=args.lr,
                  device=device,
                  img_scale=args.scale,
                  val_percent=args.val / 100,
                  amp=args.amp)
    except KeyboardInterrupt:
        torch.save(net.state_dict(), 'INTERRUPTED.pth')
        logging.info('Saved interrupt')
        raise

def get_args():
    parser = argparse.ArgumentParser(description='Train the UNet on images and target masks')
    parser.add_argument('--epochs', '-e', metavar='E', type=int, default=5, help='Number of epochs')
    parser.add_argument('--batch-size', '-b', dest='batch_size', metavar='B', type=int, default=1, help='Batch size')
    parser.add_argument('--learning-rate', '-l', metavar='LR', type=float, default=1e-5,
                        help='Learning rate', dest='lr')
    parser.add_argument('--load', '-f', type=str, default=False, help='Load model from a .pth file')
    parser.add_argument('--scale', '-s', type=float, default=0.5, help='Downscaling factor of the images')
    parser.add_argument('--validation', '-v', dest='val', type=float, default=10.0,
                        help='Percent of the data that is used as validation (0-100)')
    parser.add_argument('--amp', action='store_true', default=False, help='Use mixed precision')
    parser.add_argument('--bilinear', action='store_true', default=False, help='Use bilinear upsampling')
    parser.add_argument('--classes', '-c', type=int, default=2, help='Number of classes')

    return parser.parse_args()


if __name__ == '__main__':
    args = get_args()

    logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
    run(opt=args)




import os, sys
from pathlib import Path    
here = Path(__file__).parent
try:
    import hepai
except:
    sys.path.insert(0, f'{here.parent.parent.parent}')
    import hepai
from worker import WorkerWarper
from base_worker_model import BaseWorkerModel
from dataclasses import dataclass, field


class WorkerModel(BaseWorkerModel):
    def __init__(self, name, **kwargs):
        self.name = name  # name属性用于用于请求指定调研的模型

    @BaseWorkerModel.auto_stream  # 自动将各种类型的输出转为流式输
    def inference(self, **kwargs):
        # 自己的执行逻辑, 例如: # 
        input = kwargs.pop('input', None)
        output = [1, 2, 3, 4, 5]  # 修改为自己的输出
        return output, input
        # for i in output:
        #     yield i  # 可以return返回python的基础类型或yield生成器
    
    def chat_completions(self, **kwargs):
        pass
        raise NotImplementedError("Please implement this method")

def run_worker(model_args, worker_args, **kwargs):
    
    model = WorkerModel(  # 获取模型
        name=model_args.name
        # 此处可以传入其他参数
        )
    
    if worker_args.test:
        ret = model.inference(input='test')
        print(ret)
        return

    
    # hai.worker.start(
    WorkerWarper.start(
        # daemon=True,  # 是否以守护进程的方式启动
        model=model,
        worker_args=worker_args,
        **kwargs
        )

# (1) 实现WorkerModel
@dataclass
class ModelArgs:
    name: str = "hepai/demo_worker"  # worker的名称，用于注册到控制器
    # 其他参数

# (2) worker的参数配置和启动代码
# 用dataclasses修饰器快速定义参数类
@dataclass
class WorkerArgs:
    host: str = "0.0.0.0"  # worker的地址，0.0.0.0表示外部可访问，127.0.0.1表示只有本机可访问
    port: str = "auto"  # worker的端口，默认从42902开始自动分配
    controller_address: str = "http://aiapi.ihep.ac.cn:42901"  # 控制器的地址
    worker_address: str = "auto"  # 默认是http://<ip>:<port>
    limit_model_concurrency: int = 5  # 限制模型的并发请求
    stream_interval: float = 0.  # 额外的流式响应间隔
    no_register: bool = False  # 不注册到控制器
    permissions: str = 'groups: PAYG'  # 模型的权限授予，分为用户和组，用;分隔，例如：需要授权给所有组、a用户、b用户：'groups: all; users: a, b; owner: c'
    description: str = 'This is a demo worker in HeiAI-Distributed Deploy Framework'  # 模型的描述
    author: str = 'hepai'  # 模型的作者
    test: bool = False  # 测试模式，不会真正启动worker，只会打印参数

if __name__ == '__main__':
    # worker_args = hai.parse_args_into_dataclasses(WorkerArgs)  # 解析参数
    model_args, worker_args = hepai.parse_args_into_dataclasses((ModelArgs, WorkerArgs))  # 解析多个参数类
    # print(worker_args)
    worker_args.controller_address = "http://ainpu.ihep.ac.cn:21601"
    run_worker(model_args=model_args, worker_args=worker_args)




import os
from hepai import HepAI
from hepai import RemoteModel
    
def connect(
    name: str = "hepai/markitdown",
    base_url: str = "https://aiapi.ihep.ac.cn/apiv2",
    api_key: str = os.getenv("HEPAI_API_KEY"),
    **kwargs,
):
    """
    Connect to hepai resources. such as remote model, llm, etc.
    Args:
        name (str): The name of the resource to connect to.
        base_url (str): The base URL of the resource.
    """
    client = HepAI(
        base_url=base_url,
        api_key=api_key,
        **kwargs,
        )
    
    model: RemoteModel = client.get_remote_model(model_name=name)
    
    return model



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable, Optional
from typing_extensions import Literal

import httpx

from .... import _legacy_response
from ...._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from ...._utils import (
    maybe_transform,
    async_maybe_transform,
)
from ...._compat import cached_property
from .checkpoints import (
    Checkpoints,
    AsyncCheckpoints,
    CheckpointsWithRawResponse,
    AsyncCheckpointsWithRawResponse,
    CheckpointsWithStreamingResponse,
    AsyncCheckpointsWithStreamingResponse,
)
from ...._resource import SyncAPIResource, AsyncAPIResource
from ...._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ....pagination import SyncCursorPage, AsyncCursorPage
from ...._base_client import (
    AsyncPaginator,
    make_request_options,
)
from ....types.fine_tuning import job_list_params, job_create_params, job_list_events_params
from ....types.fine_tuning.fine_tuning_job import FineTuningJob
from ....types.fine_tuning.fine_tuning_job_event import FineTuningJobEvent

__all__ = ["Jobs", "AsyncJobs"]


class Jobs(SyncAPIResource):
    @cached_property
    def checkpoints(self) -> Checkpoints:
        return Checkpoints(self._client)

    @cached_property
    def with_raw_response(self) -> JobsWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return the
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/openai/openai-python#accessing-raw-response-data-eg-headers
        """
        return JobsWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> JobsWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/openai/openai-python#with_streaming_response
        """
        return JobsWithStreamingResponse(self)

    def create(
        self,
        *,
        model: Union[str, Literal["babbage-002", "davinci-002", "gpt-3.5-turbo", "gpt-4o-mini"]],
        training_file: str,
        hyperparameters: job_create_params.Hyperparameters | NotGiven = NOT_GIVEN,
        integrations: Optional[Iterable[job_create_params.Integration]] | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        suffix: Optional[str] | NotGiven = NOT_GIVEN,
        validation_file: Optional[str] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> FineTuningJob:
        """
        Creates a fine-tuning job which begins the process of creating a new model from
        a given dataset.

        Response includes details of the enqueued job including job status and the name
        of the fine-tuned models once complete.

        [Learn more about fine-tuning](https://platform.openai.com/docs/guides/fine-tuning)

        Args:
          model: The name of the model to fine-tune. You can select one of the
              [supported models](https://platform.openai.com/docs/guides/fine-tuning#which-models-can-be-fine-tuned).

          training_file: The ID of an uploaded file that contains training data.

              See [upload file](https://platform.openai.com/docs/api-reference/files/create)
              for how to upload a file.

              Your dataset must be formatted as a JSONL file. Additionally, you must upload
              your file with the purpose `fine-tune`.

              The contents of the file should differ depending on if the model uses the
              [chat](https://platform.openai.com/docs/api-reference/fine-tuning/chat-input) or
              [completions](https://platform.openai.com/docs/api-reference/fine-tuning/completions-input)
              format.

              See the [fine-tuning guide](https://platform.openai.com/docs/guides/fine-tuning)
              for more details.

          hyperparameters: The hyperparameters used for the fine-tuning job.

          integrations: A list of integrations to enable for your fine-tuning job.

          seed: The seed controls the reproducibility of the job. Passing in the same seed and
              job parameters should produce the same results, but may differ in rare cases. If
              a seed is not specified, one will be generated for you.

          suffix: A string of up to 64 characters that will be added to your fine-tuned model
              name.

              For example, a `suffix` of "custom-model-name" would produce a model name like
              `ft:gpt-4o-mini:openai:custom-model-name:7p4lURel`.

          validation_file: The ID of an uploaded file that contains validation data.

              If you provide this file, the data is used to generate validation metrics
              periodically during fine-tuning. These metrics can be viewed in the fine-tuning
              results file. The same data should not be present in both train and validation
              files.

              Your dataset must be formatted as a JSONL file. You must upload your file with
              the purpose `fine-tune`.

              See the [fine-tuning guide](https://platform.openai.com/docs/guides/fine-tuning)
              for more details.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/fine_tuning/jobs",
            body=maybe_transform(
                {
                    "model": model,
                    "training_file": training_file,
                    "hyperparameters": hyperparameters,
                    "integrations": integrations,
                    "seed": seed,
                    "suffix": suffix,
                    "validation_file": validation_file,
                },
                job_create_params.JobCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=FineTuningJob,
        )

    def retrieve(
        self,
        fine_tuning_job_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> FineTuningJob:
        """
        Get info about a fine-tuning job.

        [Learn more about fine-tuning](https://platform.openai.com/docs/guides/fine-tuning)

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not fine_tuning_job_id:
            raise ValueError(f"Expected a non-empty value for `fine_tuning_job_id` but received {fine_tuning_job_id!r}")
        return self._get(
            f"/fine_tuning/jobs/{fine_tuning_job_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=FineTuningJob,
        )

    def list(
        self,
        *,
        after: str | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> SyncCursorPage[FineTuningJob]:
        """
        List your organization's fine-tuning jobs

        Args:
          after: Identifier for the last job from the previous pagination request.

          limit: Number of fine-tuning jobs to retrieve.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/fine_tuning/jobs",
            page=SyncCursorPage[FineTuningJob],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "after": after,
                        "limit": limit,
                    },
                    job_list_params.JobListParams,
                ),
            ),
            model=FineTuningJob,
        )

    def cancel(
        self,
        fine_tuning_job_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> FineTuningJob:
        """
        Immediately cancel a fine-tune job.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not fine_tuning_job_id:
            raise ValueError(f"Expected a non-empty value for `fine_tuning_job_id` but received {fine_tuning_job_id!r}")
        return self._post(
            f"/fine_tuning/jobs/{fine_tuning_job_id}/cancel",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=FineTuningJob,
        )

    def list_events(
        self,
        fine_tuning_job_id: str,
        *,
        after: str | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> SyncCursorPage[FineTuningJobEvent]:
        """
        Get status updates for a fine-tuning job.

        Args:
          after: Identifier for the last event from the previous pagination request.

          limit: Number of events to retrieve.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not fine_tuning_job_id:
            raise ValueError(f"Expected a non-empty value for `fine_tuning_job_id` but received {fine_tuning_job_id!r}")
        return self._get_api_list(
            f"/fine_tuning/jobs/{fine_tuning_job_id}/events",
            page=SyncCursorPage[FineTuningJobEvent],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "after": after,
                        "limit": limit,
                    },
                    job_list_events_params.JobListEventsParams,
                ),
            ),
            model=FineTuningJobEvent,
        )


class AsyncJobs(AsyncAPIResource):
    @cached_property
    def checkpoints(self) -> AsyncCheckpoints:
        return AsyncCheckpoints(self._client)

    @cached_property
    def with_raw_response(self) -> AsyncJobsWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return the
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/openai/openai-python#accessing-raw-response-data-eg-headers
        """
        return AsyncJobsWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncJobsWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/openai/openai-python#with_streaming_response
        """
        return AsyncJobsWithStreamingResponse(self)

    async def create(
        self,
        *,
        model: Union[str, Literal["babbage-002", "davinci-002", "gpt-3.5-turbo", "gpt-4o-mini"]],
        training_file: str,
        hyperparameters: job_create_params.Hyperparameters | NotGiven = NOT_GIVEN,
        integrations: Optional[Iterable[job_create_params.Integration]] | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        suffix: Optional[str] | NotGiven = NOT_GIVEN,
        validation_file: Optional[str] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> FineTuningJob:
        """
        Creates a fine-tuning job which begins the process of creating a new model from
        a given dataset.

        Response includes details of the enqueued job including job status and the name
        of the fine-tuned models once complete.

        [Learn more about fine-tuning](https://platform.openai.com/docs/guides/fine-tuning)

        Args:
          model: The name of the model to fine-tune. You can select one of the
              [supported models](https://platform.openai.com/docs/guides/fine-tuning#which-models-can-be-fine-tuned).

          training_file: The ID of an uploaded file that contains training data.

              See [upload file](https://platform.openai.com/docs/api-reference/files/create)
              for how to upload a file.

              Your dataset must be formatted as a JSONL file. Additionally, you must upload
              your file with the purpose `fine-tune`.

              The contents of the file should differ depending on if the model uses the
              [chat](https://platform.openai.com/docs/api-reference/fine-tuning/chat-input) or
              [completions](https://platform.openai.com/docs/api-reference/fine-tuning/completions-input)
              format.

              See the [fine-tuning guide](https://platform.openai.com/docs/guides/fine-tuning)
              for more details.

          hyperparameters: The hyperparameters used for the fine-tuning job.

          integrations: A list of integrations to enable for your fine-tuning job.

          seed: The seed controls the reproducibility of the job. Passing in the same seed and
              job parameters should produce the same results, but may differ in rare cases. If
              a seed is not specified, one will be generated for you.

          suffix: A string of up to 64 characters that will be added to your fine-tuned model
              name.

              For example, a `suffix` of "custom-model-name" would produce a model name like
              `ft:gpt-4o-mini:openai:custom-model-name:7p4lURel`.

          validation_file: The ID of an uploaded file that contains validation data.

              If you provide this file, the data is used to generate validation metrics
              periodically during fine-tuning. These metrics can be viewed in the fine-tuning
              results file. The same data should not be present in both train and validation
              files.

              Your dataset must be formatted as a JSONL file. You must upload your file with
              the purpose `fine-tune`.

              See the [fine-tuning guide](https://platform.openai.com/docs/guides/fine-tuning)
              for more details.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/fine_tuning/jobs",
            body=await async_maybe_transform(
                {
                    "model": model,
                    "training_file": training_file,
                    "hyperparameters": hyperparameters,
                    "integrations": integrations,
                    "seed": seed,
                    "suffix": suffix,
                    "validation_file": validation_file,
                },
                job_create_params.JobCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=FineTuningJob,
        )

    async def retrieve(
        self,
        fine_tuning_job_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> FineTuningJob:
        """
        Get info about a fine-tuning job.

        [Learn more about fine-tuning](https://platform.openai.com/docs/guides/fine-tuning)

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not fine_tuning_job_id:
            raise ValueError(f"Expected a non-empty value for `fine_tuning_job_id` but received {fine_tuning_job_id!r}")
        return await self._get(
            f"/fine_tuning/jobs/{fine_tuning_job_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=FineTuningJob,
        )

    def list(
        self,
        *,
        after: str | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AsyncPaginator[FineTuningJob, AsyncCursorPage[FineTuningJob]]:
        """
        List your organization's fine-tuning jobs

        Args:
          after: Identifier for the last job from the previous pagination request.

          limit: Number of fine-tuning jobs to retrieve.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/fine_tuning/jobs",
            page=AsyncCursorPage[FineTuningJob],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "after": after,
                        "limit": limit,
                    },
                    job_list_params.JobListParams,
                ),
            ),
            model=FineTuningJob,
        )

    async def cancel(
        self,
        fine_tuning_job_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> FineTuningJob:
        """
        Immediately cancel a fine-tune job.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not fine_tuning_job_id:
            raise ValueError(f"Expected a non-empty value for `fine_tuning_job_id` but received {fine_tuning_job_id!r}")
        return await self._post(
            f"/fine_tuning/jobs/{fine_tuning_job_id}/cancel",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=FineTuningJob,
        )

    def list_events(
        self,
        fine_tuning_job_id: str,
        *,
        after: str | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AsyncPaginator[FineTuningJobEvent, AsyncCursorPage[FineTuningJobEvent]]:
        """
        Get status updates for a fine-tuning job.

        Args:
          after: Identifier for the last event from the previous pagination request.

          limit: Number of events to retrieve.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not fine_tuning_job_id:
            raise ValueError(f"Expected a non-empty value for `fine_tuning_job_id` but received {fine_tuning_job_id!r}")
        return self._get_api_list(
            f"/fine_tuning/jobs/{fine_tuning_job_id}/events",
            page=AsyncCursorPage[FineTuningJobEvent],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "after": after,
                        "limit": limit,
                    },
                    job_list_events_params.JobListEventsParams,
                ),
            ),
            model=FineTuningJobEvent,
        )


class JobsWithRawResponse:
    def __init__(self, jobs: Jobs) -> None:
        self._jobs = jobs

        self.create = _legacy_response.to_raw_response_wrapper(
            jobs.create,
        )
        self.retrieve = _legacy_response.to_raw_response_wrapper(
            jobs.retrieve,
        )
        self.list = _legacy_response.to_raw_response_wrapper(
            jobs.list,
        )
        self.cancel = _legacy_response.to_raw_response_wrapper(
            jobs.cancel,
        )
        self.list_events = _legacy_response.to_raw_response_wrapper(
            jobs.list_events,
        )

    @cached_property
    def checkpoints(self) -> CheckpointsWithRawResponse:
        return CheckpointsWithRawResponse(self._jobs.checkpoints)


class AsyncJobsWithRawResponse:
    def __init__(self, jobs: AsyncJobs) -> None:
        self._jobs = jobs

        self.create = _legacy_response.async_to_raw_response_wrapper(
            jobs.create,
        )
        self.retrieve = _legacy_response.async_to_raw_response_wrapper(
            jobs.retrieve,
        )
        self.list = _legacy_response.async_to_raw_response_wrapper(
            jobs.list,
        )
        self.cancel = _legacy_response.async_to_raw_response_wrapper(
            jobs.cancel,
        )
        self.list_events = _legacy_response.async_to_raw_response_wrapper(
            jobs.list_events,
        )

    @cached_property
    def checkpoints(self) -> AsyncCheckpointsWithRawResponse:
        return AsyncCheckpointsWithRawResponse(self._jobs.checkpoints)


class JobsWithStreamingResponse:
    def __init__(self, jobs: Jobs) -> None:
        self._jobs = jobs

        self.create = to_streamed_response_wrapper(
            jobs.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            jobs.retrieve,
        )
        self.list = to_streamed_response_wrapper(
            jobs.list,
        )
        self.cancel = to_streamed_response_wrapper(
            jobs.cancel,
        )
        self.list_events = to_streamed_response_wrapper(
            jobs.list_events,
        )

    @cached_property
    def checkpoints(self) -> CheckpointsWithStreamingResponse:
        return CheckpointsWithStreamingResponse(self._jobs.checkpoints)


class AsyncJobsWithStreamingResponse:
    def __init__(self, jobs: AsyncJobs) -> None:
        self._jobs = jobs

        self.create = async_to_streamed_response_wrapper(
            jobs.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            jobs.retrieve,
        )
        self.list = async_to_streamed_response_wrapper(
            jobs.list,
        )
        self.cancel = async_to_streamed_response_wrapper(
            jobs.cancel,
        )
        self.list_events = async_to_streamed_response_wrapper(
            jobs.list_events,
        )

    @cached_property
    def checkpoints(self) -> AsyncCheckpointsWithStreamingResponse:
        return AsyncCheckpointsWithStreamingResponse(self._jobs.checkpoints)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from ..._models import BaseModel

__all__ = ["TranscriptionSegment"]


class TranscriptionSegment(BaseModel):
    id: int
    """Unique identifier of the segment."""

    avg_logprob: float
    """Average logprob of the segment.

    If the value is lower than -1, consider the logprobs failed.
    """

    compression_ratio: float
    """Compression ratio of the segment.

    If the value is greater than 2.4, consider the compression failed.
    """

    end: float
    """End time of the segment in seconds."""

    no_speech_prob: float
    """Probability of no speech in the segment.

    If the value is higher than 1.0 and the `avg_logprob` is below -1, consider this
    segment silent.
    """

    seek: int
    """Seek offset of the segment."""

    start: float
    """Start time of the segment in seconds."""

    temperature: float
    """Temperature parameter used for generating the segment."""

    text: str
    """Text content of the segment."""

    tokens: List[int]
    """Array of token IDs for the text content."""



"""
把所有client模块以外从外部引用的类定义引用到这里，方便把haiddf拷贝到hepai项目中使用
"""


# from ....controller.worker_manager._worker_class import (
#     HRemoteModel, WorkerInfo, WorkerStatusInfo, WorkerStoppedInfo,
#     )

# from ....controller.user_manager._user_class import (
#     UserInfo, UserDeletedInfo, APIKeyInfo, APIKeyDeletedInfo,
#     )
# from ...worker.worker_app import HWorkerAPP

### --- 为HepAI --- ###
from ..base_class._worker_class import (
    HRemoteModel, WorkerInfo, WorkerStatusInfo, WorkerStoppedInfo,
    )

from ..base_class._user_class import (
    UserInfo, UserDeletedInfo, APIKeyInfo, APIKeyDeletedInfo,
    )

from ..base_class._agent_class import AgentInfo



"""
Worker类别，具备基本和示例功能
"""

import time
import json
import time
from typing import List, Union, Generator
import threading
import uuid
import traceback
import argparse
# import dataclasses
from dataclasses import dataclass, field
import os, sys
from pathlib import Path

from fastapi import FastAPI
from fastapi import BackgroundTasks
from fastapi import HTTPException
from fastapi import APIRouter
from fastapi import Request
from fastapi.responses import StreamingResponse, JSONResponse, Response, FileResponse
from starlette.requests import Request as StarletteRequest
import requests
import uvicorn

import damei as dm 
import atexit
here = Path(__file__).parent

try:
    from hai.uaii.worker.utils import auto_port, auto_worker_address, pretty_print_semaphore
    from hai.uaii.worker.base_worker_model import BaseWorkerModel
except:
    sys.path.append(str(here.parent.parent.parent))
    from hai.uaii.worker.utils import auto_port, auto_worker_address, pretty_print_semaphore
    from hai.uaii.worker.base_worker_model import BaseWorkerModel


GB = 1 << 30

logger = dm.get_logger('worker.py')
global_counter = 0
model_semaphore = None
WORKER_HEART_BEAT_INTERVAL = 30


def heart_beat_worker(controller):
    while True:
        time.sleep(WORKER_HEART_BEAT_INTERVAL)
        controller.send_heart_beat()


class Worker:
    """
    Worker class, with basic and example functions
    """
    def __init__(self, 
                 controller_addr: str,  # controller的地址
                 worker_addr: str,  # 本worker的地址
                 worker_id: str = None,  # worker的uuid
                 model=None,  # 模型
                 limit_model_concurrency: int = 5,  # 模型并发数
                 stream_interval: int = 2,  # stream的间隔
                 no_register: bool = False,  # 是否注册到controller
                 permissions: any = None,  # 权限
                 description: str = None,  # 模型的描述
                 author: str = None,  # 模型的作者
                 **kwargs
                 ):
        
        self.controller_addr = controller_addr
        self.worker_addr = worker_addr
        self.worker_id = worker_id if worker_id else str(uuid.uuid4())[:6]
        self.limit_model_concurrency = limit_model_concurrency
        self.stream_interval = stream_interval
        self.no_register = no_register
        self.description = description
        self.author = author
        self._permissions = self._init_permissions(permissions)  # 将str转换为dict

        self._model = model or BaseWorkerModel()

        if not no_register:
            self.register_to_controller()
            self.heart_beat_thread = threading.Thread(
                target=heart_beat_worker, args=(self,))
            self.heart_beat_thread.setDaemon(True)  # 解决推出问题
            self.heart_beat_thread.start()
        atexit.register(self.exit_handler)

    def _init_permissions(self, permissions):
        """worker授予用户或者组的权限"""
        prems = dict()
        if permissions is None:
            pass
        elif isinstance(permissions, str):
            """user: <user1>; user: <user2>; group: <group1>, ..."""
            for a in permissions.split(';'):
                user_or_group, name = a.split(':')
                user_or_group = user_or_group.strip()
                name = name.strip()
                assert user_or_group in ['owner', 'users', 'groups']
                names = name.split(',')  if ',' in name else [name]
                prems[user_or_group] = names
        elif isinstance(permissions, dict):
            prems = permissions
        else:
            raise ValueError(f"permissions should be str or dict, but got {type(permissions)}")
        return prems
    

    def check_model(self):
        # 测试是否有inference函数
        assert hasattr(self.model, 'inference'), f"Model {self.model_name} has no inference function"

    @property
    def permissions(self):
        return self._permissions
    
    @property
    def model_name(self):
        if self.model is None:
            return None
        return self.model.name

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        self._model = model  
        
    # 程序退出时，发送退出信息
    def exit_handler(self):
        """
        退出时向Contoller发送退出信息
        """
        if self.no_register:  # 初始化时没有注册到controller，不需要发送退出信息
            return

        logger.info(f'Remove model "{self.model_name}" from controller {self.controller_addr}, ')
        url = self.controller_addr + "/stop_worker"
        
        metadata = {
            "description": self.description,
            "author": self.author,
        }
        data = {
            "worker_addr": self.worker_addr,
            "check_heart_beat": True,
            "metadata": metadata,
            "worker_status": self.get_status(),
            }
        r = requests.post(url, json=data)
        assert r.status_code == 200, f"Stop model {self.model_name} failed. {r.text}"
        logger.info(f'Done. {r.text}')  
    
    def register_to_controller(self):
        logger.info(f'Register model "{self.model_name}" to controller.')
        url = self.controller_addr + "/register_worker"
        
        metadata = {
            "description": self.description,
            "author": self.author,
        }
        data = {
            "worker_addr": self.worker_addr,
            "check_heart_beat": True,
            "metadata": metadata,
            "worker_status": self.get_status(),
            }
        r = requests.post(url, json=data)
        assert r.status_code == 200, f"Register model {self.model_name} failed. {r.text}"
        
    def get_status(self):
        try:
            misc = self.model.get_misc()
        except:
            misc = None
        return {
            "model_name": self.model_name,
            "speed": 1,
            "queue_length": self.get_queue_length(),
            "permissions": self.permissions,
            "misc": misc,
        }

    def send_heart_beat(self):
        logger.info(f"Send heart beat. Models: {self.model_name}. "
                    f"Semaphore: {pretty_print_semaphore(model_semaphore)}. "
                    f"global_counter: {global_counter}")

        url = self.controller_addr + "/receive_heart_beat"

        while True:
            try:
                ret = requests.post(url, json={
                    "worker_addr": self.worker_addr,
                    "queue_length": self.get_queue_length()}, timeout=5)
                exist = ret.json()["exist"]
                break
            except requests.exceptions.RequestException as e:
                logger.error(f"heart beat error: {e}")
            time.sleep(5)

        if not exist:
            self.register_to_controller()

    def get_queue_length(self):
        if model_semaphore is None:
            return 0
        else:
            _value = model_semaphore._value
            _num_waiters = len(model_semaphore._waiters) if model_semaphore._waiters is not None else 0
            return self.limit_model_concurrency - _value + _num_waiters

    def _sleep(self, seconds):
        if float(seconds) == 0. or seconds is None:
            return
        if seconds < 0:
            raise ValueError(f"stream_interval should be positive, but got {seconds}")
        if seconds > 3:
            raise ValueError(f"stream_interval should be less than 3, but got {seconds}")
        time.sleep(seconds)

    
    def get_generator_content_and_verify_one(self, generator):
        """
        从生成器中获取内容，并验证是否只有一个元素
        """
        try:
            content = next(generator)
            print('content: ', content)
        except StopIteration as e:
            print(f'Generator is empty: {e} {type(e)}')
            return e
            # raise ValueError(f"Generator is empty: {e} {type(e)}")
        try:
            next(generator)
            raise ValueError("Generator should only have one element")
        except StopIteration:
            pass
        return content
    
    def unified_gate(self, **kwargs):
        """
        统一入口
        """
        assert "function" in kwargs, "function is required"
        function = kwargs.pop("function")
        try:
            if hasattr(self.model, function) and callable(getattr(self.model, function)):
                func = getattr(self.model, function)
                try:
                    ret = func(**kwargs)
                    return True, ret
                except Exception as e:
                    error_info = f'{type(e)} {e}'
                    # logger.error(f'error_info: {error_info}\n {traceback.format_exc()}')
                    return False, error_info
            else:
                return False, f"Function `{function}` does not exist or is not callable in the model `{self.model_name}`",
        except Exception as e:
            error_info = f'{type(e)} {e}'
            logger.error(f'error_info: {error_info}\n {traceback.format_exc()}')
            return False, error_info
    
    def generate_stream_gate(self, **params):
        """
        生成流式响应的门
        :param params:
                model: 模型名
                其他参数
        :return:
        """
        
        stream = params.get("stream", False)
        stream_interval = params.get("stream_interval", self.stream_interval)

        def yield_data(output):
            if isinstance(output, int):
                output = str(output)
                yield output.encode() + b"\0"
            elif isinstance(output, str):
                for x in output:
                    yield x.encode() + b"\0"
            elif isinstance(output, bytes):
                yield output + b"\0"
            elif isinstance(output, list):
                output = str(output)
                yield output.encode() + b"\0"
            elif isinstance(output, dict):
                yield json.dumps(output).encode() + b"\0"
            else:
                raise ValueError(f'output type {type(output)} not supported')
        try:
            output = self.model.inference(**params)
            if not stream:
                # print(f'output: {output}')
                new_output = dict()
                new_output["message"] = output
                new_output["status_code"] = 42901
                yield json.dumps(new_output).encode()
                return
            else:
                if isinstance(output, (int, str, bytes, list, dict)):
                    yield from yield_data(output)
                    self._sleep(stream_interval)
                # 如果是python的generator
                elif hasattr(output, '__next__'):
                    for x in output:
                        yield from yield_data(x)
                        self._sleep(stream_interval)
                else:
                    raise ValueError(f'output type {type(output)} not supported')
                yield "[DONE]".encode() + b"\0"

        except Exception as e:
            error_info = f'{type(e)} {e}'
            logger.error(f'error_info: {error_info}\n {traceback.format_exc()}')
            ret = {
                "message": error_info,
                "status_code": 42904, 
            }
            yield json.dumps(ret).encode() + b"\0"

    def train(self, **kwargs):
        ret = self.model.train(**kwargs)
        return ret


from fastapi import FastAPI, Request, BackgroundTasks
from fastapi.responses import StreamingResponse
import asyncio


class WorkerAPP(FastAPI):

    def __init__(self):
        super().__init__()
        self._worker: Worker = None

    @property
    def worker(self):
        if self._worker is None:
            raise ValueError(
                f"worker is not initialized, \
                please call `WorkerAPP.worker = BaseWorker(**kwargs)` first")
        return self._worker
    
    @worker.setter
    def worker(self, worker):
        self._worker = worker
        
app = WorkerAPP()  # It's a FastAPI instance


def release_model_semaphore():
    model_semaphore.release()

async def create_receive_channel(data: dict):
    async def receive():
        return {
            "type": "http.request",
            "body": json.dumps(data).encode('utf-8'),
            "more_body": False,
        }
    return receive
    

@app.post("/v1/chat/completions")   # unifiled_gate衍生出来的chat_completions接口，用于直接向worker发送补全请求
async def v1_chat_complietions(request: Request):
    original_data = await request.json()
    # 修改请求数据，添加新字段
    modified_data = original_data.copy()
    modified_data['function'] = 'chat_completions'
    # 创建一个新的receive通道，包含修改后的数据
    custom_receive = await create_receive_channel(modified_data)
    custom_request = StarletteRequest(scope=request.scope, receive=custom_receive)
    return await app_unified_gate(custom_request)

@app.post("/v1/worker/unified_gate")
async def v1_unified_gate(request: Request):
    return await app_unified_gate(request)

@app.post("/worker_unified_gate")
async def app_unified_gate(request: Request):
    global model_semaphore, global_counter
    global_counter += 1
    params = await request.json()
    if model_semaphore is None:
        model_semaphore = asyncio.Semaphore(
            app.worker.limit_model_concurrency)
    await model_semaphore.acquire()
    stream = params.get("stream", False)
    ok, data = app.worker.unified_gate(**params)
    if not ok:
        data = data if isinstance(data, str) else json.dumps(data)
        # print(model_semaphore)
        release_model_semaphore()
        raise HTTPException(status_code=404, detail=data)
    background_tasks = BackgroundTasks()  # 背景任务
    background_tasks.add_task(release_model_semaphore)  # 释放锁
    if isinstance(data, Generator) and (stream is None or stream is False):
        logger.info(f"Get a generator from function: {params.get('function')}, auto set stream=True")
        stream = True
    if not stream:
        # 判断是不是文件路径
        if isinstance(data, Path):
            if not os.path.exists(data):
                raise HTTPException(status_code=404, detail=f"File not found: {data}")
            return FileResponse(data, filename=data.name, background=background_tasks)
        return JSONResponse(content=data, background=background_tasks)
    return StreamingResponse(data, background=background_tasks)


@app.post("/worker_generate_stream")
async def generate_stream(request: Request):
    global model_semaphore, global_counter
    global_counter += 1
    params = await request.json()

    if model_semaphore is None:
        model_semaphore = asyncio.Semaphore(
            app.worker.limit_model_concurrency)
        # 用于设置模型的并发请求，并发达到该值时，后续请求会被阻塞，直到之前的请求完成释放了锁
    await model_semaphore.acquire()
    generator = app.worker.generate_stream_gate(**params)
    background_tasks = BackgroundTasks()  # 背景任务
    background_tasks.add_task(release_model_semaphore)  # 释放锁
    return StreamingResponse(generator, background=background_tasks)

@app.post("/train")
async def train(request: Request):
    global model_semaphore, global_counter
    global_counter += 1
    params = await request.json()
    if model_semaphore is None:
        model_semaphore = asyncio.Semaphore(
            app.worker.limit_model_concurrency
        )
    await model_semaphore.acquire()
    result = app.worker.train(**params)
    background_tasks = BackgroundTasks()  # 背景任务
    background_tasks.add_task(release_model_semaphore)  # 释放锁
    return result


@app.post("/worker_get_status")
async def get_status(request: Request):
    return app.worker.get_status()


def run_worker(model=None, worker_args=None, extra_routers=None, daemon=False, test=False, **kwargs):
    args = worker_args or WorkerArgs()
    # print(f'worker args: {args}')
    if args.port == "auto":
        start_port = args.__dict__.get("start_port", 42902)
        args.port = auto_port(args.port, start=start_port)
    args.worker_address = auto_worker_address(args.worker_address, args.host, args.port)
    logger.info(f"Worker address: {args.worker_address}")
    if test:
        args.controller_address = "http://chat.ihep.ac.cn:4444"

    # 更新
    for k, v in kwargs.items():
        if hasattr(args, k):
            setattr(args, k, v)

    worker  = Worker(
        args.controller_address,
        args.worker_address,
        model=model,
        limit_model_concurrency=args.limit_model_concurrency,
        stream_interval=args.stream_interval,
        no_register=args.no_register,
        permissions=args.permissions,
        description=args.description,
        author=args.author,
        )
    
    app.worker = worker
    if extra_routers:
        if isinstance(extra_routers, APIRouter):
            extra_routers = [extra_routers]
        for extra_router in extra_routers:
            assert isinstance(extra_router, APIRouter), f"extra_router should be an instance of APIRouter: {extra_router}"
            app.include_router(extra_router)

    logger.info(f"Controller address: {args.controller_address}")

    if daemon:
        logger.info(f"Daemon Worker address: {args.worker_address}")
        import threading
        t = threading.Thread(target=uvicorn.run, args=(app,), kwargs={
            "host": args.host, "port": args.port, "log_level": "info"
            })
        t.setDaemon(True)
        t.start()
    else:
        uvicorn.run(app, host=args.host, port=args.port, log_level="info")
    return args.worker_address


@dataclass
class WorkerArgs:
    host: str = "0.0.0.0"  # worker的地址，0.0.0.0表示外部可访问，127.0.0.1表示只有本机可访问
    port: str = "auto"  # 默认从42902开始
    controller_address: str = "http://aiapi.ihep.ac.cn:42901"  # 控制器的地址
    worker_address: str = "auto"  # 默认是http://<ip>:<port>
    limit_model_concurrency: int = 5  # 限制模型的并发请求
    stream_interval: float = 0.  # 额外的流式响应间隔
    no_register: bool = False  # 不注册到控制器
    permissions: str = 'groups: all'  # 模型的权限授予，分为用户和组，用;分隔
    description: str = None  # 模型的描述
    author: str = None  # 模型的作者


class WorkerWarper:

    @staticmethod
    def start(**kwargs):
        """
        运行一个hepai的worker
        :param model: BaseWorkerModel = None, 模型
        :param worker_args: WorkerArgs = None, worker的参数，可以由WorkerArgs类生成，会被下面的参数覆盖
        :param host: str = "0.0.0.0"  # worker的地址，0.0.0.0表示外部可访问，127.0.0.1表示只有本机可访问
        :param port: str = "auto"  # 默认从42902开始
        :param extra_routers: List[APIRouter] = None,  # 额外的路由
        :param controller_address: str = "http://chat.ihep.ac.cn:42901"  # 控制器的地址
        :param worker_address: str = "auto"  # 默认是http://<ip>:<port>
        :param limit_model_concurrency: int = 5  # 限制模型的并发请求
        :param stream_interval: float = 0.  # 额外的流式响应间隔
        :param no_register: bool = False  # 不注册到控制器
        :param permissions: str = 'group: all'  # 模型的权限授予，分为用户和组，用;分隔
        """
        daemon = kwargs.pop('daemon', False)
        assert daemon is False, "Daemon is deprecated in 2023.10.12"
        return run_worker(daemon=daemon, **kwargs)
    
    


if __name__ == "__main__":
    run_worker(test=False)
    
    









# -*- coding=utf-8 -*-

"""
 Verion: 1.0
 Since : 2.7.18
 Author: zhangjian
 Site: https://github.com/xarrow/
 File: timer.py
 Time: 2020/8/30
 
 Add New Functional nacos-sdk-python
"""

import threading


class NacosTimer(object):
    __slots__ = ['_name', '_timer', '_fn', '_interval', '_ignore_ex', '_on_result', '_on_exception',
                 '_args', '_kwargs']

    def __init__(self,
                 name,
                 fn,
                 interval=7,
                 *args,
                 **kwargs):
        """
        NacosTimer
        :param name:  timer name
        :param fn:    function which scheduler
        :param interval:  scheduler interval, default 7s
        :param args:      args in function
        :param kwargs:    kwargs in function
        """
        #
        self._name = name
        # Thread.Timer
        self._timer = None
        # function which callable
        self._fn = fn
        # timer interval default 7s
        self._interval = interval
        # whether ignore invoke exception
        self._ignore_ex = False
        self._on_result = None
        self._on_exception = None
        # function args
        self._args = args
        # function kwargs
        self._kwargs = kwargs

    @property
    def name(self):
        return self._name

    def set_name(self, name):
        self._name = name
        return self

    @property
    def fn(self):
        return self._fn

    def set_fn(self, fn):
        self._fn = fn
        return self

    @property
    def interval(self, ):
        return self._interval

    def set_interval(self, interval):
        self._interval = interval
        return self

    @property
    def ignore_ex(self):
        return self._ignore_ex

    def set_ignore_ex(self, ignore_ex):
        self._ignore_ex = ignore_ex
        return self

    @property
    def on_result(self):
        return self._on_result

    def set_on_result(self, fn):
        self._on_result = fn
        return self

    @property
    def on_exception(self):
        return self._on_exception

    def set_on_exception(self, fn):
        self._on_exception = fn
        return self

    def alive(self):
        if self._timer is None:
            return False
        return self._timer.is_alive()

    def scheduler(self):
        try:
            res = self._fn(*self._args, **self._kwargs)
            if self._on_result:
                self._on_result(res)
        except Exception as ex:
            if self._on_exception:
                self._on_exception(ex)
            if not self._ignore_ex:
                # stop timer
                raise ex
        self._timer = threading.Timer(self._interval, self.scheduler, )
        self._timer.start()

    def cancel(self):
        if self._timer:
            self._timer.cancel()


class NacosTimerManager(object):
    def __init__(self, ):
        self._timers_container = {}
        self._executed = False

    def all_timers(self):
        return self._timers_container

    def add_timer(self, timer):
        self._timers_container[timer.name] = timer
        return self

    def execute(self):
        """
        scheduler all timer in manager
        :return: None
        """
        if self._executed:
            return
        for name, timer in self._timers_container.items():
            if timer.alive():
                continue
            timer.scheduler()
        self._executed = True

    def cancel_timer(self, timer_name=None, ):
        """
        cancel timer , and  nacos timer still in container
        it can execute again.
        :param timer_name:
        :return: None
        """
        timer = self._timers_container.get(timer_name)
        if timer:
            timer.cancel()

    def cancel(self):
        """
        cancel all timer in container
        :return: None
        """
        for _, timer in self._timers_container.items():
            timer.cancel()

    def stop_timer(self, timer_name):
        """
        cancel nacos timer and remove it from timer container
        :param timer_name:
        :return: None
        """
        self.cancel_timer(timer_name)
        self._timers_container.pop(timer_name)

    def stop(self):
        """
        remove all timer, and it can not execute again
        """
        self.cancel()
        self._timers_container.clear()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hello.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0bhello.proto\"3\n\x0cHelloRequest\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\t\x12\x15\n\x05skill\x18\x02 \x01(\x0b\x32\x06.Skill\"\x84\x01\n\rHelloResponse\x12\x0e\n\x06result\x18\x01 \x01(\t\x12\x31\n\nmap_result\x18\x02 \x03(\x0b\x32\x1d.HelloResponse.MapResultEntry\x1a\x30\n\x0eMapResultEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\"\x15\n\x05Skill\x12\x0c\n\x04name\x18\x01 \x01(\t27\n\x0bGrpcService\x12(\n\x05hello\x12\r.HelloRequest\x1a\x0e.HelloResponse\"\x00\x42\x03\x80\x01\x01\x62\x06proto3')



_HELLOREQUEST = DESCRIPTOR.message_types_by_name['HelloRequest']
_HELLORESPONSE = DESCRIPTOR.message_types_by_name['HelloResponse']
_HELLORESPONSE_MAPRESULTENTRY = _HELLORESPONSE.nested_types_by_name['MapResultEntry']
_SKILL = DESCRIPTOR.message_types_by_name['Skill']
HelloRequest = _reflection.GeneratedProtocolMessageType('HelloRequest', (_message.Message,), {
  'DESCRIPTOR' : _HELLOREQUEST,
  '__module__' : 'hello_pb2'
  # @@protoc_insertion_point(class_scope:HelloRequest)
  })
_sym_db.RegisterMessage(HelloRequest)

HelloResponse = _reflection.GeneratedProtocolMessageType('HelloResponse', (_message.Message,), {

  'MapResultEntry' : _reflection.GeneratedProtocolMessageType('MapResultEntry', (_message.Message,), {
    'DESCRIPTOR' : _HELLORESPONSE_MAPRESULTENTRY,
    '__module__' : 'hello_pb2'
    # @@protoc_insertion_point(class_scope:HelloResponse.MapResultEntry)
    })
  ,
  'DESCRIPTOR' : _HELLORESPONSE,
  '__module__' : 'hello_pb2'
  # @@protoc_insertion_point(class_scope:HelloResponse)
  })
_sym_db.RegisterMessage(HelloResponse)
_sym_db.RegisterMessage(HelloResponse.MapResultEntry)

Skill = _reflection.GeneratedProtocolMessageType('Skill', (_message.Message,), {
  'DESCRIPTOR' : _SKILL,
  '__module__' : 'hello_pb2'
  # @@protoc_insertion_point(class_scope:Skill)
  })
_sym_db.RegisterMessage(Skill)

_GRPCSERVICE = DESCRIPTOR.services_by_name['GrpcService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\200\001\001'
  _HELLORESPONSE_MAPRESULTENTRY._options = None
  _HELLORESPONSE_MAPRESULTENTRY._serialized_options = b'8\001'
  _HELLOREQUEST._serialized_start=15
  _HELLOREQUEST._serialized_end=66
  _HELLORESPONSE._serialized_start=69
  _HELLORESPONSE._serialized_end=201
  _HELLORESPONSE_MAPRESULTENTRY._serialized_start=153
  _HELLORESPONSE_MAPRESULTENTRY._serialized_end=201
  _SKILL._serialized_start=203
  _SKILL._serialized_end=224
  _GRPCSERVICE._serialized_start=226
  _GRPCSERVICE._serialized_end=281
# @@protoc_insertion_point(module_scope)




import httpx
# from .openai_api.adapted_openai._exceptions import APIStatusError
from .openai_api import APIStatusError




class HAPIStatusError(APIStatusError):
    response: httpx.Response
    status_code: int
    request_id: str | None

    def __init__(self, message: str, *, response: httpx.Response, body: object | None) -> None:
        super().__init__(message, response=response, body=body)
        
    def __repr__(self):
        return f'{self.__class__.__name__}(message={self.message!r}, status_code={self.status_code!r}, request_id={self.request_id!r})'

    def __str__(self):
        # 只显示类名和消息
        return f'{self.__class__.__name__}: {self.args[0]}'

